"""Module for calculating and plotting the timestamp differences.

This script utilizes an unpacking module used specifically for the
LinoSPAD2 binary data output.

This file can also be imported as a module and contains the following
functions:

    TODO remove
    * calculate_and_save_timestamp_differences - unpacks the binary data,
    calculates timestamp differences, and saves into a '.feather' file.
    Works with firmware versions '2208' and '2212b'.

    * calculate_and_save_timestamp_differences_fast - unpacks the binary
    data, calculates timestamp differences, and saves into a '.feather'
    file. Works with firmware versions '2208' and '2212b'. Uses a faster
    algorithm than the function above.

    * calculate_and_save_timestamp_differences_full_sensor - unpacks the
    binary data, calculates timestamp differences and saves into a
    '.feather' file. Works with firmware versions '2208', '2212s' and
    '2212b'. Analyzes data from both sensor halves/both FPGAs. Useful
    for data where the signal is at 0 across the whole sensor from the
    start of data collecting.

    * calculate_and_save_timestamp_differences_full_sensor_alt - unpacks
    the binary data, calculates timestamp differences and saves into a
    '.feather' file. Works with firmware versions '2208', '2212s' and
    '2212b'. Analyzes data from both sensor halves/both FPGAs. Useful
    for data where signal is above zero right from the start.

    * collect_and_plot_timestamp_differences - collect timestamps from a
    '.feather' file and plot them in a grid.

    TODO remove
    * collect_and_plot_timestamp_differences_from_ft_file - unpack the 
    '.feather' file requested, collect timestamp differences and
    plot it as a grid of plots. Useful when only the '.feather' file
    is available and not the raw '.dat' files.

    * collect_and_plot_timestamp_differences_full_sensor - collect
    timestamps from a '.feather' file and plot histograms of them
    in a grid. This function should be used for the full sensor
    setup.
"""

import glob
import os
import sys
from math import ceil
from typing import List
from warnings import warn

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pyarrow import feather as ft
from tqdm import tqdm

from LinoSPAD2.functions import calc_diff as cd
from LinoSPAD2.functions import unpack as f_up
from LinoSPAD2.functions import utils


def _flatten(input_list: List):
    """Flatten the input list.

    Flatten the input list, which can be a list of numbers, lists,
    or a combination of the two above, and return a list of
    numbers only, unpacking the lists inside.

    Parameters
    ----------
    input_list : List
        Input list that can contain numbers, lists of numbers, or a
        combination of both.

    Returns
    -------
    list
        Flattened list of numbers only.
    """
    flattened = []
    for item in input_list:
        if isinstance(item, list):
            flattened.extend(item)
        else:
            flattened.append(item)
    return flattened


def calculate_and_save_timestamp_differences(
    path: str,
    pixels: List[int] | List[List[int]],
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    delta_window: float = 50e3,
    app_mask: bool = True,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
    correct_pix_address: bool = False,
):
    """Calculate and save timestamp differences into '.feather' file.

    Unpacks data into a dictionary, calculates timestamp differences for
    the requested pixels, and saves them into a '.feather' table. Works with
    firmware version 2212.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files.
    pixels : List[int] | List[List[int]]
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    rewrite : bool
        switch for rewriting the plot if it already exists. used as a
        safeguard to avoid unwanted overwriting of the previous results.
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the '#'.
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while apply_offset_calibration is set to 'False', only the TDC
        calibration is applied. The default is True.
    absolute_timestamps: bool, optional
        Indicator for data with absolute timestamps. The default is
        False.
    correct_pix_address : bool, optional
        Correct pixel address for the sensor half on side 23 of the
        daughterboard. The default is False.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'daughterboard_number', 'motherboard_number', and 'firmware_version'
        are accepted. The first error is raised so that the plot does not
        accidentally get rewritten in the case no clear input was given.

    Returns
    -------
    None.

    Examples
    -------
    For the sensor half on the '23' side of the daughterboard, the
    pixel addressing should be correct. Let's assume the offset
    calibration was not done for this sensor and, therefore, the
    calibration matrix is not available - it should be passed as False.
    Let's collect timestamp differences for pairs of pixels 15-25,
    15-26, and 15-27.

    First, get the absolute path to where the '.dat' files are.
    >>> path = r'C:/Path/To/Data'

    Now to the function itself.
    >>> calculate_and_save_timestamp_differences(
    >>> path,
    >>> pixels = [15, [25,26,27]],
    >>> rewrite = True,
    >>> daughterboard_number="NL11",
    >>> motherboard_number="#21",
    >>> firmware_version="2212s",
    >>> timestamps = 1000,
    >>> include_offset = False,
    >>> correct_pixel_addressing = True,
    >>> )
    """

    # TODO: remove
    warn(
        "This function is deprecated. Use"
        "'calculate_and_save_timestamp_differences_fast'",
        DeprecationWarning,
        stacklevel=2,
    )

    # Parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or '2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")

    os.chdir(path)

    # Handle the input pixel list
    pixels = utils.pixel_list_transform(pixels)

    files_all = glob.glob("*.dat*")
    files_all.sort(key=os.path.getmtime)

    out_file_name = files_all[0][:-4] + "-" + files_all[-1][:-4]

    # Feather file counter for saving delta ts into separate files
    # of up to 100 MB
    ft_file_number = 0

    # Check if the feather file exists and if it should be rewrited
    feather_file = os.path.join(
        path, "delta_ts_data", f"{out_file_name}.feather"
    )

    # Remove the old '.feather' files with the pattern
    # for ft_file in feather_files:
    utils.file_rewrite_handling(feather_file, rewrite)

    # # Go back to the folder with '.dat' files
    # os.chdir(path)

    # Collect the data for the required pixels
    print(
        "\n> > > Collecting data for delta t plot for the requested "
        "pixels and saving it to .feather in a cycle < < <\n"
    )
    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    # Correct pixel addressing for motherboard on side '23'
    if correct_pix_address:
        pixels = utils.correct_pixels_address(pixels)

    # Mask the hot/warm pixels
    if app_mask is True:
        mask = utils.apply_mask(daughterboard_number, motherboard_number)
        if isinstance(pixels[0], int) and isinstance(pixels[1], int):
            pixels = [pix for pix in pixels if pix not in mask]
        else:
            pixels[0] = [pix for pix in pixels[0] if pix not in mask]
            pixels[1] = [pix for pix in pixels[1] if pix not in mask]

    for i in tqdm(range(ceil(len(files_all))), desc="Collecting data"):
        file = files_all[i]

        # Prepare a dictionary for output
        deltas_all = {}

        # Unpack data for the requested pixels into dictionary
        if not absolute_timestamps:
            data_all = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            data_all, _ = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )

        # Calculate the timestamp differences for the given pixels
        deltas_all = cd.calculate_differences_2212(
            data_all, pixels, pix_coor, delta_window
        )

        # Save data as a .feather file in a cycle so data is not lost
        # in the case of failure close to the end
        data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
        del deltas_all
        data_for_plot_df = data_for_plot_df.T
        try:
            os.chdir("delta_ts_data")
        except FileNotFoundError:
            os.mkdir("delta_ts_data")
            os.chdir("delta_ts_data")

        # Check if feather file exists
        feather_file = f"{out_file_name}_{ft_file_number}.feather"
        if os.path.isfile(feather_file):
            # Check the size of the existing '.feather', if larger
            # than 100 MB, create new one
            if os.path.getsize(feather_file) / 1024 / 1024 < 100:
                # Load existing feather file
                existing_data = ft.read_feather(feather_file)

                # Append new data to the existing feather file
                combined_data = pd.concat(
                    [existing_data, data_for_plot_df], axis=0
                )
                ft.write_feather(combined_data, feather_file)
            else:
                ft_file_number += 1
                feather_file = f"{out_file_name}_{ft_file_number}.feather"
                ft.write_feather(data_for_plot_df, feather_file)

        else:
            # Save as a new feather file
            ft.write_feather(data_for_plot_df, feather_file)
        os.chdir("..")

    # Combine the numbered feather files into a single one
    utils.combine_feather_files(path)

    # Check, if the file was created
    if (
        os.path.isfile(path + f"/delta_ts_data/{out_file_name}.feather")
        is True
    ):
        print(
            "\n> > > Timestamp differences are saved as"
            f"{out_file_name}.feather in "
            f"{os.path.join(path, 'delta_ts_data')} < < <"
        )

    else:
        print("File wasn't generated. Check input parameters.")


def calculate_and_save_timestamp_differences_fast(
    path: str,
    pixels: List[int] | List[List[int]],
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    delta_window: float = 50e3,
    cycle_length: float = None,
    app_mask: bool = True,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
    correct_pix_address: bool = False,
):
    """Calculate and save timestamp differences into '.feather' file.

    Unpacks data into a dictionary, calculates timestamp differences for
    the requested pixels, and saves them into a '.feather' table. Works with
    firmware version 2212. Uses a faster algorithm.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files.
    pixels : List[int] | List[List[int]]
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    rewrite : bool
        switch for rewriting the plot if it already exists. used as a
        safeguard to avoid unwanted overwriting of the previous results.
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the '#'.
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    cycle_length: float, optional
        Length of the acquisition cycle. The default is None.
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while apply_offset_calibration is set to 'False', only the TDC
        calibration is applied. The default is True.
    absolute_timestamps: bool, optional
        Indicator for data with absolute timestamps. The default is
        False.
    correct_pix_address : bool, optional
        Correct pixel address for the sensor half on side 23 of the
        daughterboard. The default is False.

    Raises
    ------
    TypeError
        Raised if "pixels" is not a list.
    TypeError
        Raised if "firmware_version" is not a string.
    TypeError
        Raised if "rewrite" is not a boolean.
    TypeError
        Raised if "daughterboard_number" is not a string.
    """
    # Parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or '2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")

    os.chdir(path)

    # Handle the input list
    pixels = utils.pixel_list_transform(pixels)
    files_all = glob.glob("*.dat")

    files_all = sorted(files_all)

    out_file_name = files_all[0][:-4] + "-" + files_all[-1][:-4]

    # Feather file counter for saving delta ts into separate files
    # of up to 100 MB
    ft_file_number = 0

    # Check if the feather file exists and if it should be rewrited
    feather_file = os.path.join(
        path, "delta_ts_data", f"{out_file_name}.feather"
    )

    # Remove the old '.feather' files with the pattern
    # for ft_file in feather_files:
    utils.file_rewrite_handling(feather_file, rewrite)

    # Go back to the folder with '.dat' files
    os.chdir(path)

    # Collect the data for the required pixels
    print(
        "\n> > > Collecting data for delta t plot for the requested "
        "pixels and saving it to .feather in a cycle < < <\n"
    )
    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    # Correct pixel addressing for motherboard on side '23'
    if correct_pix_address:
        pixels = utils.correct_pixels_address(pixels)

    # Mask the hot/warm pixels
    if app_mask is True:
        mask = utils.apply_mask(daughterboard_number, motherboard_number)
        if isinstance(pixels[0], int) and isinstance(pixels[1], int):
            pixels = [pix for pix in pixels if pix not in mask]
        else:
            pixels[0] = [pix for pix in pixels[0] if pix not in mask]
            pixels[1] = [pix for pix in pixels[1] if pix not in mask]

    for i in tqdm(range(ceil(len(files_all))), desc="Collecting data"):
        file = files_all[i]

        # Unpack data for the requested pixels into dictionary
        if not absolute_timestamps:
            data_all = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            data_all, _ = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )

        # If cycle_length is not given manually, estimate from the data
        if cycle_length is None:
            cycle_length = np.max(data_all)

        delta_ts = cd.calculate_differences_2212_fast(
            data_all, pixels, pix_coor, delta_window, cycle_length
        )

        # Save data as a .feather file in a cycle so data is not lost
        # in the case of failure close to the end
        delta_ts = pd.DataFrame.from_dict(delta_ts, orient="index")
        delta_ts = delta_ts.T

        try:
            os.chdir("delta_ts_data")
        except FileNotFoundError:
            os.mkdir("delta_ts_data")
            os.chdir("delta_ts_data")

        # Check if feather file exists
        feather_file = f"{out_file_name}_{ft_file_number}.feather"
        if os.path.isfile(feather_file):
            # Check the size of the existing '.feather', if larger
            # than 100 MB, create new one
            if os.path.getsize(feather_file) / 1024 / 1024 < 10:
                # Load existing feather file
                existing_data = ft.read_feather(feather_file)

                # Append new data to the existing feather file
                combined_data = pd.concat([existing_data, delta_ts], axis=0)
                ft.write_feather(combined_data, feather_file)
            else:
                ft_file_number += 1
                feather_file = f"{out_file_name}_{ft_file_number}.feather"
                ft.write_feather(delta_ts, feather_file)

        else:
            # Save as a new feather file
            ft.write_feather(delta_ts, feather_file)
        os.chdir("..")

    # Combine the numbered feather files into a single one
    utils.combine_feather_files(path)

    # Check, if the file was created
    if (
        os.path.isfile(path + f"/delta_ts_data/{out_file_name}.feather")
        is True
    ):
        print(
            "\n> > > Timestamp differences are saved as"
            f"{out_file_name}.feather in "
            f"{os.path.join(path, 'delta_ts_data')} < < <"
        )

    else:
        print("File wasn't generated. Check input parameters.")


def calculate_and_save_timestamp_differences_full_sensor(
    path,
    pixels: list,
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number1: str,
    motherboard_number2: str,
    firmware_version: str,
    timestamps: int = 512,
    delta_window: float = 50e3,
    app_mask: bool = True,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
):  # TODO add option for collecting from more than just two pixels
    # TODO use pixel handling function for modularity (if possible)
    """Calculate and save timestamp differences into '.feather' file.

    Unpacks data into a dictionary, calculates timestamp differences for
    the requested pixels and saves them into a '.feather' table. Works with
    firmware version 2212. Analyzes data from both sensor halves/both
    FPGAs, hence the two input parameters for LinoSPAD2 motherboards.

    Parameters
    ----------
    path : str
        Path to where two folders with data from both motherboards
        are. The folders should be named after the motherboards.
    pixels : list
        List of two pixels, one from each sensor half.
    rewrite : bool
        switch for rewriting the plot if it already exists. used as a
        safeguard to avoid unwanted overwriting of the previous results.
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number1 : str
        LinoSPAD2 motherboard (FPGA) number, including the '#'.
    motherboard_number2 : str
        Second LinoSPAD2 motherboard (FPGA) number.
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC calibration is
        applied. The default is True.
    absolute_timestamps : bool, optional
        Switch for unpacking data that were collected together with
        absolute timestamps. The default is False.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'daughterboard_number', 'motherboard_number', and 'firmware_version'
        are accepted. The first error is raised so that the plot does not
        accidentally get rewritten in the case no clear input was given.
    FileNotFoundError
        Raised if data from the first LinoSPAD2 motherboard were not
        found.
    FileNotFoundError
        Raised if data from the second LinoSPAD2 motherboard were not
        found.
    """
    # parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or" "'2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")

    os.chdir(path)

    # Check the data from the first FPGA board
    try:
        os.chdir(f"{motherboard_number1}")
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            f"Data from {motherboard_number1} not found"
        ) from exc
    # files_all1 = sorted(glob.glob("*.dat*"))
    files_all1 = glob.glob("*.dat*")
    files_all1.sort(key=os.path.getmtime)
    out_file_name = files_all1[0][:-4]
    os.chdir("..")

    # Check the data from the second FPGA board
    try:
        os.chdir(f"{motherboard_number2}")
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            f"Data from {motherboard_number2} not found"
        ) from exc
    # files_all2 = sorted(glob.glob("*.dat*"))
    files_all2 = glob.glob("*.dat*")
    files_all2.sort(key=os.path.getmtime)
    out_file_name = out_file_name + "-" + files_all2[-1][:-4]
    os.chdir("..")

    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    # Check if '.feather' file with timestamps differences already
    # exists

    feather_file = os.path.join(
        path, "delta_ts_data", f"{out_file_name}.feather"
    )

    utils.file_rewrite_handling(feather_file, rewrite)

    # TODO add check for masked/noisy pixels
    # if app_mask is True:
    #     path_to_back = os.getcwd()
    #     path_to_mask = os.path.realpath(__file__) + "/../.." + "/params/masks"
    #     os.chdir(path_to_mask)
    #     file_mask1 = glob.glob("*{}_{}*".format(daughterboard_number, motherboard_number1))[0]
    #     mask1 = np.genfromtxt(file_mask1).astype(int)
    #     file_mask2 = glob.glob("*{}_{}*".format(daughterboard_number, motherboard_number2))[0]
    #     mask2 = np.genfromtxt(file_mask2).astype(int)
    #     os.chdir(path_to_back)

    abs_tmsp_list1 = []
    abs_tmsp_list2 = []
    for i in tqdm(range(ceil(len(files_all1))), desc="Collecting data"):
        deltas_all = {}

        # First board, unpack data
        os.chdir(f"{motherboard_number1}")
        file = files_all1[i]
        if not absolute_timestamps:
            data_all1 = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number1,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            (
                data_all1,
                abs_tmsp1,
            ) = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number1,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        # abs_tmsp_list1.append(abs_tmsp1)
        # Collect indices of cycle ends (the '-2's)
        cycle_ends1 = np.where(data_all1[0].T[1] == -2)[0]
        cyc1 = np.argmin(
            np.abs(cycle_ends1 - np.where(data_all1[:].T[1] > 0)[0].min())
        )
        if cycle_ends1[cyc1] > np.where(data_all1[:].T[1] > 0)[0].min():
            cyc1 = cyc1 - 1
            cycle_start1 = cycle_ends1[cyc1]
        else:
            cycle_start1 = cycle_ends1[cyc1]

        os.chdir("..")

        # Second board, unpack data
        os.chdir(f"{motherboard_number2}")
        file = files_all2[i]
        if not absolute_timestamps:
            data_all2 = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number2,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            (
                data_all2,
                abs_tmsp2,
            ) = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number2,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        # abs_tmsp_list2.append(abs_tmsp2)
        # Collect indices of cycle ends (the '-2's)
        cycle_ends2 = np.where(data_all2[0].T[1] == -2)[0]
        cyc2 = np.argmin(
            np.abs(cycle_ends2 - np.where(data_all2[:].T[1] > 0)[0].min())
        )
        if cycle_ends2[cyc2] > np.where(data_all2[:].T[1] > 0)[0].min():
            cyc2 = cyc2 - 1
            cycle_start2 = cycle_ends2[cyc2]
        else:
            cycle_start2 = cycle_ends2[cyc2]

        if cyc1 > cyc2:
            abs_tmsp_list1.append(abs_tmsp1[cyc1:])
            abs_tmsp_list2.append(abs_tmsp2[cyc2 : -cyc1 + cyc2])
        else:
            abs_tmsp_list1.append(abs_tmsp1[cyc1 : -cyc2 + cyc1])
            abs_tmsp_list2.append(abs_tmsp2[cyc2:])

        os.chdir("..")

        pix_left_peak = pixels[0]
        # The following piece of code take any value for the pixel from
        # the second motherboard, either in terms of full sensor (so
        # a value >=256) or in terms of single sensor half

        if pixels[1] >= 256:
            if pixels[1] > 256 + 127:
                pix_right_peak = 255 - (pixels[1] - 256)
            else:
                pix_right_peak = pixels[1] - 256 + 128
        elif pixels[1] > 127:
            pix_right_peak = 255 - pixels[1]
        else:
            pix_right_peak = pixels[1] + 128

        # Get the data from the requested pixel only
        deltas_all[f"{pix_left_peak},{pix_right_peak}"] = []
        tdc1, pix_c1 = np.argwhere(pix_coor == pix_left_peak)[0]
        pix1 = np.where(data_all1[tdc1].T[0] == pix_c1)[0]
        tdc2, pix_c2 = np.argwhere(pix_coor == pix_right_peak)[0]
        pix2 = np.where(data_all2[tdc2].T[0] == pix_c2)[0]

        # Data from one of the board should be shifted as data collection
        # on one of the board is started later
        if cycle_start1 > cycle_start2:
            cyc = len(data_all1[0].T[1]) - cycle_start1 + cycle_start2
            cycle_ends1 = cycle_ends1[cycle_ends1 >= cycle_start1]
            cycle_ends2 = np.intersect1d(
                cycle_ends2[cycle_ends2 >= cycle_start2],
                cycle_ends2[cycle_ends2 <= cyc],
            )

        else:
            cyc = len(data_all1[0].T[1]) - cycle_start2 + cycle_start1
            cycle_ends2 = cycle_ends2[cycle_ends2 >= cycle_start2]
            cycle_ends1 = np.intersect1d(
                cycle_ends1[cycle_ends1 >= cycle_start1],
                cycle_ends1[cycle_ends1 <= cyc],
            )

        # Get timestamps for both pixels in the given cycle
        for cyc in range(len(cycle_ends1) - 1):
            pix1_ = pix1[
                np.logical_and(
                    pix1 >= cycle_ends1[cyc], pix1 < cycle_ends1[cyc + 1]
                )
            ]
            if not np.any(pix1_):
                continue
            pix2_ = pix2[
                np.logical_and(
                    pix2 >= cycle_ends2[cyc], pix2 < cycle_ends2[cyc + 1]
                )
            ]

            if not np.any(pix2_):
                continue
            # Calculate delta t
            tmsp1 = data_all1[tdc1].T[1][
                pix1_[np.where(data_all1[tdc1].T[1][pix1_] > 0)[0]]
            ]
            tmsp2 = data_all2[tdc2].T[1][
                pix2_[np.where(data_all2[tdc2].T[1][pix2_] > 0)[0]]
            ]
            for t1 in tmsp1:
                deltas = tmsp2 - t1
                ind = np.where(np.abs(deltas) < delta_window)[0]
                deltas_all[f"{pix_left_peak},{pix_right_peak}"].extend(
                    deltas[ind]
                )

        # Save data to a feather file in a cycle so data is not lost
        # in the case of failure close to the end
        data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
        del deltas_all
        data_for_plot_df = data_for_plot_df.T

        feather_file = f"{out_file_name}.feather"

        try:
            os.chdir("delta_ts_data")
        except FileNotFoundError:
            os.mkdir("delta_ts_data")
            os.chdir("delta_ts_data")

        if os.path.isfile(feather_file):
            # Load existing Feather file
            existing_data = ft.read_feather(feather_file)

            # Append new data to the existing Feather file
            combined_data = pd.concat(
                [existing_data, data_for_plot_df], axis=0
            )
            ft.write_feather(combined_data, feather_file)

        else:
            # Save as a new Feather file
            ft.write_feather(data_for_plot_df, feather_file)

        os.chdir("..")

    # Check if the file with the results was created
    if (
        os.path.isfile(os.path.join(path, f"/delta_ts_data/{feather_file}"))
        is True
    ):
        print(
            "\n> > > Timestamp differences are saved as"
            f"{feather_file} in {os.path.join(path, 'delta_ts_data/')} < < <"
        )
    else:
        print("File wasn't generated. Check input parameters.")

    return abs_tmsp_list1, abs_tmsp_list2


def calculate_and_save_timestamp_differences_full_sensor_alt(
    path,
    pixels: list,
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number1: str,
    motherboard_number2: str,
    firmware_version: str,
    timestamps: int = 512,
    delta_window: float = 50e3,
    threshold: int = 0,
    app_mask: bool = True,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
):  # TODO add option for collecting from more than just two pixels
    # TODO use pixel handling function for modularity (if possible)
    """Calculate and save timestamp differences into '.feather' file.

    Unpacks data into a dictionary, calculates timestamp differences for
    the requested pixels and saves them into a '.feather' table. Works with
    firmware version 2212. Analyzes data from both sensor halves/both
    FPGAs, hence the two input parameters for LinoSPAD2 motherboards.
    Uses the threshold value to find the first cycle in each sensor half
    where the signal is above that value. Useful for when the signal
    is above zero right from the start of data collecting.

    Parameters
    ----------
    path : str
        Path to where two folders with data from both motherboards
        are. The folders should be named after the motherboards.
    pixels : list
        List of two pixels, one from each sensor half.
    rewrite : bool
        switch for rewriting the plot if it already exists. used as a
        safeguard to avoid unwanted overwriting of the previous results.
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number1 : str
        First LinoSPAD2 motherboard (FPGA) number, including the '#'.
    motherboard_number2 : str
        Second LinoSPAD2 motherboard (FPGA) number, including the '#'.
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    threshold: int, optional
        Threshold for the number of timestamps per cycle in the given
        pixels that is used to find the specific cycle. With value of 0
        this will find the first cycle where there is any positive signal
        in the pixel, while a value of 15 will find first cycle when
        signal is above approx. 4 kHz strong. The default is 0.
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC calibration is
        applied. The default is True.
    absolute_timestamps : bool, optional
        Switch for unpacking data that were collected together with
        absolute timestamps. The default is False.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'daughterboard_number', 'motherboard_number', and 'firmware_version'
        are accepted. The first error is raised so that the plot does not
        accidentally get rewritten in the case no clear input was given.
    FileNotFoundError
        Raised if data from the first LinoSPAD2 motherboard were not
        found.
    FileNotFoundError
        Raised if data from the second LinoSPAD2 motherboard were not
        found.
    """
    # parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or" "'2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")

    os.chdir(path)

    # Check the data from the first FPGA board
    try:
        os.chdir(f"{motherboard_number1}")
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            f"Data from {motherboard_number1} not found"
        ) from exc
    files_all1 = glob.glob("*.dat*")
    files_all1.sort(key=lambda x: os.path.getmtime(x))
    out_file_name = files_all1[0][:-4]
    os.chdir("..")

    # Check the data from the second FPGA board
    try:
        os.chdir(f"{motherboard_number2}")
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            f"Data from {motherboard_number2} not found"
        ) from exc
    files_all2 = glob.glob("*.dat*")
    files_all2.sort(key=lambda x: os.path.getmtime(x))
    out_file_name = out_file_name + "-" + files_all2[-1][:-4]
    os.chdir("..")

    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    # Check if '.feather' file with timestamps differences already
    # exists
    feather_file = os.path.join(
        path, "delta_ts_data", f"{out_file_name}.feather"
    )

    utils.file_rewrite_handling(feather_file, rewrite)

    # TODO add check for masked/noisy pixels
    # if app_mask is True:
    #     path_to_back = os.getcwd()
    #     path_to_mask = os.path.realpath(__file__) + "/../.." + "/params/masks"
    #     os.chdir(path_to_mask)
    #     file_mask1 = glob.glob("*{}_{}*".format(daughterboard_number, motherboard_number1))[0]
    #     mask1 = np.genfromtxt(file_mask1).astype(int)
    #     file_mask2 = glob.glob("*{}_{}*".format(daughterboard_number, motherboard_number2))[0]
    #     mask2 = np.genfromtxt(file_mask2).astype(int)
    #     os.chdir(path_to_back)

    abs_tmsp_list1 = []
    abs_tmsp_list2 = []
    for i in tqdm(range(ceil(len(files_all1))), desc="Collecting data"):
        deltas_all = {}

        # First board, unpack data
        os.chdir(f"{motherboard_number1}")
        file = files_all1[i]
        if not absolute_timestamps:
            data_all1 = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number1,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            (
                data_all1,
                abs_tmsp1,
            ) = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number1,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )

        os.chdir("..")

        # Second board, unpack data
        os.chdir(f"{motherboard_number2}")
        file = files_all2[i]
        if not absolute_timestamps:
            data_all2 = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number2,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            (
                data_all2,
                abs_tmsp2,
            ) = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number2,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )

        pix_left_peak = pixels[0]

        # The following piece of code take any value for the pixel from
        # the second motherboard, either in terms of full sensor (so
        # a value >=256) or in terms of single sensor half
        if pixels[1] >= 256:
            if pixels[1] > 256 + 127:
                pix_right_peak = 255 - (pixels[1] - 256)
            else:
                pix_right_peak = pixels[1] - 256 + 128
        elif pixels[1] > 127:
            pix_right_peak = 255 - pixels[1]
        else:
            pix_right_peak = pixels[1] + 128

        # Get the data from the requested pixel only
        deltas_all[f"{pix_left_peak},{pix_right_peak}"] = []
        tdc1, pix_c1 = np.argwhere(pix_coor == pix_left_peak)[0]
        pix1 = np.where(data_all1[tdc1].T[0] == pix_c1)[0]
        tdc2, pix_c2 = np.argwhere(pix_coor == pix_right_peak)[0]
        pix2 = np.where(data_all2[tdc2].T[0] == pix_c2)[0]

        cycle_ends1 = np.where(data_all1[0].T[1] == -2)[0]
        cycle_ends2 = np.where(data_all2[0].T[1] == -2)[0]

        ends1 = np.insert(cycle_ends1, 0, 0)

        # Timestamps only from the chosen TDCs
        column_data1 = data_all1[tdc1].T[1]
        column_data2 = data_all2[tdc2].T[1]

        # For number of timestamps per cycle for the chosen pixels
        pixel_cycle_pop1 = []
        pixel_cycle_pop2 = []

        # Populate the lists
        for i in range(len(ends1) - 1):
            cycle_indices1 = pix1[(pix1 >= ends1[i]) & (pix1 < ends1[i + 1])]
            cycle_indices2 = pix2[(pix2 >= ends1[i]) & (pix2 < ends1[i + 1])]
            pixel_cycle_pop1.append(
                len(
                    column_data1[cycle_indices1][
                        column_data1[cycle_indices1] > 0
                    ]
                )
            )
            pixel_cycle_pop2.append(
                len(
                    column_data2[cycle_indices2][
                        column_data2[cycle_indices2] > 0
                    ]
                )
            )

        # Find the starting cycle based on the value of threshold
        cycle_start_index1 = np.where(np.array(pixel_cycle_pop1) > threshold)[
            0
        ].min()
        cycle_start_index2 = np.where(np.array(pixel_cycle_pop2) > threshold)[
            0
        ].min()

        cycle_start1 = cycle_ends1[cycle_start_index1]
        cycle_start2 = cycle_ends2[cycle_start_index2]

        # Cut the absolute timestamps so that they have the same length
        if absolute_timestamps:
            if cycle_start_index1 > cycle_start_index2:
                abs_tmsp_list1.append(abs_tmsp1[cycle_start_index1:])
                abs_tmsp_list2.append(
                    abs_tmsp2[
                        cycle_start_index2 : -cycle_start_index1
                        + cycle_start_index2
                    ]
                )
            else:
                abs_tmsp_list1.append(
                    abs_tmsp1[
                        cycle_start_index1 : -cycle_start_index2
                        + cycle_start_index1
                    ]
                )
                abs_tmsp_list2.append(abs_tmsp2[cycle_start_index2:])

        os.chdir("..")

        # Data from one of the board should be shifted as data collection
        # on one of the board is started later
        if cycle_start1 > cycle_start2:
            cyc = len(data_all1[0].T[1]) - cycle_start1 + cycle_start2
            cycle_ends1 = cycle_ends1[cycle_ends1 >= cycle_start1]
            cycle_ends2 = np.intersect1d(
                cycle_ends2[cycle_ends2 >= cycle_start2],
                cycle_ends2[cycle_ends2 < cyc],
            )

        else:
            cyc = len(data_all1[0].T[1]) - cycle_start2 + cycle_start1
            cycle_ends2 = cycle_ends2[cycle_ends2 >= cycle_start2]
            cycle_ends1 = np.intersect1d(
                cycle_ends1[cycle_ends1 >= cycle_start1],
                cycle_ends1[cycle_ends1 < cyc],
            )

        # Get timestamps for both pixels in the given cycle
        for cyc in range(len(cycle_ends1) - 1):
            pix1_ = pix1[
                np.logical_and(
                    pix1 >= cycle_ends1[cyc], pix1 < cycle_ends1[cyc + 1]
                )
            ]
            if not np.any(pix1_):
                continue
            pix2_ = pix2[
                np.logical_and(
                    pix2 >= cycle_ends2[cyc], pix2 < cycle_ends2[cyc + 1]
                )
            ]

            if not np.any(pix2_):
                continue
            # Calculate delta t
            tmsp1 = data_all1[tdc1].T[1][
                pix1_[np.where(data_all1[tdc1].T[1][pix1_] > 0)[0]]
            ]
            tmsp2 = data_all2[tdc2].T[1][
                pix2_[np.where(data_all2[tdc2].T[1][pix2_] > 0)[0]]
            ]
            for t1 in tmsp1:
                deltas = tmsp2 - t1
                ind = np.where(np.abs(deltas) < delta_window)[0]
                deltas_all[f"{pix_left_peak},{pix_right_peak}"].extend(
                    deltas[ind]
                )

        # Save data to a feather file in a cycle so data is not lost
        # in the case of failure close to the end
        data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
        del deltas_all
        data_for_plot_df = data_for_plot_df.T

        feather_file = f"{out_file_name}.feather"

        try:
            os.chdir("delta_ts_data")
        except FileNotFoundError:
            os.mkdir("delta_ts_data")
            os.chdir("delta_ts_data")

        if os.path.isfile(feather_file):
            # Load existing Feather file
            existing_data = ft.read_feather(feather_file)

            # Append new data to the existing Feather file
            combined_data = pd.concat(
                [existing_data, data_for_plot_df], axis=0
            )
            ft.write_feather(combined_data, feather_file)

        else:
            # Save as a new Feather file
            ft.write_feather(data_for_plot_df, feather_file)

        os.chdir("..")

    # Check if the file with the results was created
    if (
        os.path.isfile(os.path.join(path, f"/delta_ts_data/{feather_file}"))
        is True
    ):
        print(
            "\n> > > Timestamp differences are saved as"
            f"{feather_file} in {os.path.join(path, 'delta_ts_data/')} < < <"
        )
    else:
        print("File wasn't generated. Check input parameters.")

    return abs_tmsp_list1, abs_tmsp_list2 if absolute_timestamps else None


def collect_and_plot_timestamp_differences(
    path,
    pixels,
    rewrite: bool,
    ft_file: str = None,
    range_left: int = -10e3,
    range_right: int = 10e3,
    multiplier: int = 1,
    same_y: bool = False,
    color: str = "rebeccapurple",
    correct_pix_address: bool = False,
):
    """Collect and plot timestamp differences from a '.feather' file.

    Plots timestamp differences from a '.feather' file as a grid of histograms
    and as a single plot. The plot is saved in the 'results/delta_t' folder,
    which is created (if it does not already exist) in the same folder
    where data are.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files.
    pixels : list
        List of pixel numbers for which the timestamp differences
        should be plotted.
    rewrite : bool
        switch for rewriting the plot if it already exists. used as a
        safeguard to avoid unwanted overwriting of the previous results.
        Switch for rewriting the plot if it already exists. Used as a
        safeguard to avoid unwanted overwriting of the previous results.
    ft_file : str, optional
        Path to the feather file with timestamp differences. If used,
        the data files in the path are ignored. The default is None.
    range_left : int, optional
        Lower limit for timestamp differences, lower values are not used.
        The default is -10e3.
    range_right : int, optional
        Upper limit for timestamp differences, higher values are not used.
        The default is 10e3.
    multiplier : int, optional
        Histogram binning multiplier. Can be used for coarser binning.
        The default is 1.
    same_y : bool, optional
        Switch for plotting the histograms with the same y-axis.
        The default is False.
    color : str, optional
        Color for the plot. The default is 'rebeccapurple'.
    correct_pix_address : bool, optional
        Correct pixel address for the sensor half on side 23 of the
        daughterboard. The default is False.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' are accepted. The error is
        raised so that the plot does not accidentally gets rewritten.

    Returns
    -------
    None.
    """
    # parameter type check
    if isinstance(rewrite, bool) is not True:
        raise TypeError("'rewrite' should be boolean")
    os.chdir(path)

    if ft_file is not None:
        feather_file_name = ft_file.split(".")[0]
    else:
        files_all = glob.glob("*.dat*")
        files_all.sort(key=os.path.getmtime)
        feather_file_name = files_all[0][:-4] + "-" + files_all[-1][:-4]

        # Check if plot exists and if it should be rewritten
        try:
            os.chdir("results/delta_t")
            if os.path.isfile(f"{feather_file_name}_delta_t_grid.png"):
                if rewrite is True:
                    print(
                        "\n! ! ! Plot of timestamp differences already"
                        "exists and will be rewritten ! ! !\n"
                    )
                else:
                    sys.exit(
                        "\nPlot already exists, 'rewrite' set to 'False', exiting."
                    )
            os.chdir("../..")
        except FileNotFoundError:
            pass

    print(
        "\n> > > Plotting timestamps differences as a grid of histograms < < <"
    )

    plt.rcParams.update({"font.size": 27})
    # In the case two lists given - the left and right peaks - _flatten
    # into a single list

    # Save to use in the title
    # pixels_title = np.copy(pixels)

    if correct_pix_address:
        for i, pixel in enumerate(pixels):
            if pixel > 127:
                pixels[i] = 255 - pixels[i]
            else:
                pixels[i] = pixels[i] + 128

    pixels = _flatten(pixels)

    if len(pixels) > 2:
        fig, axs = plt.subplots(
            len(pixels) - 1,
            len(pixels) - 1,
            figsize=(5.5 * len(pixels), 5.5 * len(pixels)),
        )
        for ax in axs:
            for x in ax:
                x.axes.set_axis_off()
    else:
        fig = plt.figure(figsize=(16, 10))

    # Check if the y limits of all plots should be the same
    if same_y is True:
        y_max_all = 0

    for q, _ in tqdm(enumerate(pixels), desc="Row in plot"):
        for w, _ in enumerate(pixels):
            if w <= q:
                continue
            if len(pixels) > 2:
                axs[q][w - 1].axes.set_axis_on()

            # Read data from Feather file
            if ft_file is not None:
                try:
                    data_to_plot = ft.read_feather(
                        ft_file, columns=[f"{pixels[q]},{pixels[w]}"]
                    ).dropna()
                except ValueError:
                    continue
            else:
                try:
                    data_to_plot = ft.read_feather(
                        f"delta_ts_data/{feather_file_name}.feather",
                        columns=[f"{pixels[q]},{pixels[w]}"],
                    ).dropna()
                except ValueError:
                    continue

            # Prepare the data for the plot
            data_to_plot = np.array(data_to_plot)
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot < range_left)
            )
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot > range_right)
            )
            # data_to_plot = data_to_plot + 1e3

            # Bins should be in units of 17.857 ps - average bin width
            # of the LinoSPAD2 TDCs
            try:
                bins = np.arange(
                    np.min(data_to_plot),
                    np.max(data_to_plot),
                    2500 / 140 * multiplier,
                )
            except ValueError:
                print(
                    f"\nCouldn't calculate bins for {q}-{w} pair: "
                    "probably not enough delta ts."
                )
                continue

            if len(pixels) > 2:
                axs[q][w - 1].set_xlabel("\u0394t (ps)")
                axs[q][w - 1].set_ylabel("# of coincidences (-)")
                n, b, p = axs[q][w - 1].hist(
                    data_to_plot,
                    bins=bins,
                    color=color,
                )
            else:
                plt.xlabel("\u0394t (ps)")
                plt.ylabel("# of coincidences (-)")
                n, b, p = plt.hist(
                    data_to_plot,
                    bins=bins,
                    color=color,
                )

            # Find number of timestamps differences in a 2 ns window
            # around the peak (HBT or cross-talk)
            try:
                peak_max_pos = np.argmax(n).astype(np.intc)
                # 2 ns window around peak
                win = int(1000 / ((range_right - range_left) / 100))
                peak_max = np.sum(n[peak_max_pos - win : peak_max_pos + win])
            except ValueError:
                peak_max = 0

            if same_y is True:
                try:
                    y_max = np.max(n)
                except ValueError:
                    y_max = 0
                    print("\nCould not find maximum y value\n")
                if y_max_all < y_max:
                    y_max_all = y_max
                if len(pixels) > 2:
                    axs[q][w - 1].set_ylim(0, y_max + 4)
                else:
                    plt.ylim(0, y_max + 4)

            if len(pixels) > 2:
                axs[q][w - 1].set_xlim(range_left - 100, range_right + 100)
                axs[q][w - 1].set_title(
                    f"Pixels {pixels[q]},{pixels[w]}"
                    # f"Pixels {pixels[q]},{pixels[w]}\nPeak in 2 ns "
                    # f"window: {int(peak_max)}"
                )
            else:
                plt.xlim(range_left - 100, range_right + 100)
                # Cut the first x tick label to avoid overlapping with
                # y-axis ticks
                ax = plt.gca()
                ticks = ax.get_xticks()
                tick_labels = ax.get_xticklabels()
                ax.set_xticks(ticks[2:-1], tick_labels[2:-1])

                plt.title(f"Pixels {pixels[q]},{pixels[w]}")

            # Save the figure
            try:
                os.chdir("results/delta_t")
            except FileNotFoundError:
                os.makedirs("results/delta_t")
                os.chdir("results/delta_t")
            fig.tight_layout()  # for perfect spacing between the plots
            plt.savefig(f"{feather_file_name}_delta_t_grid.png")
            os.chdir("../..")

    print(
        "\n> > > Plot is saved as {file} in {path}< < <".format(
            file=feather_file_name + "_delta_t_grid.png",
            path=path + "/results/delta_t",
        )
    )


def collect_and_plot_timestamp_differences_from_ft_file(
    ft_file,
    pixels,
    rewrite: bool,
    range_left: int = -10e3,
    range_right: int = 10e3,
    multiplier: int = 1,
    same_y: bool = False,
    color: str = "rebeccapurple",
):
    """Collect and plot timestamp differences from a '.feather' file.

    Plots timestamp differences from a '.feather' file as a grid of histograms
    and as a single plot. The plot is saved in the 'results/delta_t' folder,
    which is created (if it does not already exist) in the same folder
    where data are.

    Parameters
    ----------
    ft_file : str
        Absolute path to the feather file.
    pixels : list
        List of pixel numbers for which the timestamp differences
        should be plotted.
    rewrite : bool
        switch for rewriting the plot if it already exists. used as a
        safeguard to avoid unwanted overwriting of the previous results.
        Switch for rewriting the plot if it already exists. Used as a
        safeguard to avoid unwanted overwriting of the previous results.
    range_left : int, optional
        Lower limit for timestamp differences, lower values are not used.
        The default is -10e3.
    range_right : int, optional
        Upper limit for timestamp differences, higher values are not used.
        The default is 10e3.
    multiplier : int, optional
        Histogram binning multiplier. Can be used for coarser binning.
        The default is 1.
    same_y : bool, optional
        Switch for plotting the histograms with the same y-axis.
        The default is False.
    color : str, optional
        Color for the plot. The default is 'rebeccapurple'.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' are accepted. The error is
        raised so that the plot does not accidentally gets rewritten.

    Returns
    -------
    None.
    """

    # TODO: remove
    warn(
        "This function is deprecated. Use"
        "'collect_and_plot_timestamp_differences' with"
        "the 'ft_file' parameter instead",
        DeprecationWarning,
        stacklevel=2,
    )

    # parameter type check
    if isinstance(rewrite, bool) is not True:
        raise TypeError("'rewrite' should be boolean")
    # plt.ioff()

    path = os.path.dirname(ft_file)

    feather_file_name = ft_file.split("\\")[-1].split(".")[0]

    os.chdir(path)

    # Check if plot exists and if it should be rewritten
    try:
        os.chdir("results/delta_t")
        if os.path.isfile(f"{feather_file_name}_delta_t_grid.png"):
            if rewrite is True:
                print(
                    "\n! ! ! Plot of timestamp differences already"
                    "exists and will be rewritten ! ! !\n"
                )
            else:
                sys.exit(
                    "\nPlot already exists, 'rewrite' set to 'False', exiting."
                )
        os.chdir("../..")
    except FileNotFoundError:
        pass

    print(
        "\n> > > Plotting timestamps differences as a grid of histograms < < <"
    )

    plt.rcParams.update({"font.size": 27})
    # In the case two lists given - the left and right peaks - _flatten
    # into a single list
    pixels = _flatten(pixels)

    # Prepare the grid for the plots based on the number of pixels
    # given
    if len(pixels) > 2:
        fig, axs = plt.subplots(
            len(pixels) - 1,
            len(pixels) - 1,
            figsize=(5.5 * len(pixels), 5.5 * len(pixels)),
        )
        for ax in axs:
            for x in ax:
                x.axes.set_axis_off()
    else:
        fig = plt.figure(figsize=(16, 16))

    # Check if the y limits of all plots should be the same
    if same_y is True:
        y_max_all = 0

    for q, _ in tqdm(enumerate(pixels), desc="Row in plot"):
        for w, _ in enumerate(pixels):
            if w <= q:
                continue
            if len(pixels) > 2:
                axs[q][w - 1].axes.set_axis_on()

            # Read data from Feather file
            try:
                data_to_plot = ft.read_feather(
                    ft_file,
                    columns=[f"{pixels[q]},{pixels[w]}"],
                ).dropna()
            except ValueError:
                continue

            # Prepare the data for the plot
            data_to_plot = np.array(data_to_plot)
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot < range_left)
            )
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot > range_right)
            )

            # Bins should be in units of 17.857 ps - average bin width
            # of the LinoSPAD2 TDCs
            try:
                bins = np.arange(
                    np.min(data_to_plot),
                    np.max(data_to_plot),
                    2500 / 140 * multiplier,
                )
            except ValueError:
                print(
                    f"\nCouldn't calculate bins for {q}-{w} pair: "
                    "probably not enough delta ts."
                )
                continue

            if len(pixels) > 2:
                axs[q][w - 1].set_xlabel("\u0394t (ps)")
                axs[q][w - 1].set_ylabel("# of coincidences (-)")
                n, b, p = axs[q][w - 1].hist(
                    data_to_plot,
                    bins=bins,
                    color=color,
                )
            else:
                plt.xlabel("\u0394t (ps)")
                plt.ylabel("# of coincidences (-)")
                n, b, p = plt.hist(
                    data_to_plot,
                    bins=bins,
                    color=color,
                )

            # Find number of timestamps differences in a 2 ns window
            # around the peak (HBT or cross-talk)
            try:
                peak_max_pos = np.argmax(n).astype(np.intc)
                # 2 ns window around peak
                win = int(1000 / ((range_right - range_left) / 100))
                peak_max = np.sum(n[peak_max_pos - win : peak_max_pos + win])
            except ValueError:
                peak_max = 0

            if same_y is True:
                try:
                    y_max = np.max(n)
                except ValueError:
                    y_max = 0
                    print("\nCould not find maximum y value\n")
                if y_max_all < y_max:
                    y_max_all = y_max
                if len(pixels) > 2:
                    axs[q][w - 1].set_ylim(0, y_max + 4)
                else:
                    plt.ylim(0, y_max + 4)

            if len(pixels) > 2:
                axs[q][w - 1].set_xlim(range_left - 100, range_right + 100)
                axs[q][w - 1].set_title(
                    f"Pixels {pixels[q]},{pixels[w]}\nPeak in 2 ns "
                    f"window: {int(peak_max)}"
                )
            else:
                plt.xlim(range_left - 100, range_right + 100)

                plt.title(f"Pixels {pixels[q]},{pixels[w]}")

            # Save the figure
            try:
                os.chdir("results/delta_t")
            except FileNotFoundError:
                os.makedirs("results/delta_t")
                os.chdir("results/delta_t")
            fig.tight_layout()  # for perfect spacing between the plots
            plt.savefig(f"{feather_file_name}_delta_t_grid.png")
            os.chdir("../..")

    print(
        "\n> > > Plot is saved as {file} in {path}< < <".format(
            file=feather_file_name + "_delta_t_grid.png",
            path=path + "/results/delta_t",
        )
    )


def collect_and_plot_timestamp_differences_full_sensor(
    path,
    pixels,
    rewrite: bool,
    range_left: int = -10e3,
    range_right: int = 10e3,
    multiplier: int = 1,
    same_y: bool = False,
    color: str = "rebeccapurple",
):
    """Collect and plot timestamp differences from a '.feather' file.

    Plots timestamp differences from a '.feather' file as a grid of
    histograms and as a single plot. The plot is saved in the
    'results/delta_t' folder, which is created (if it does not already
    exist) in the same folder where data are.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files.
    pixels : list
        List of pixel numbers for which the timestamp differences
        should be plotted.
    rewrite : bool
        switch for rewriting the plot if it already exists. used as a
        safeguard to avoid unwanted overwriting of the previous results.
        Switch for rewriting the plot if it already exists. Used as a
        safeguard to avoid unwanted overwriting of the previous results.
    range_left : int, optional
        Lower limit for timestamp differences, lower values are not used.
        The default is -10e3.
    range_right : int, optional
        Upper limit for timestamp differences, higher values are not used.
        The default is 10e3.
    multiplier : int, optional
        Histogram binning multiplier. Can be used for coarser binning.
        The default is 1.
    same_y : bool, optional
        Switch for plotting the histograms with the same y-axis.
        The default is False.
    color : str, optional
        Color for the plot. The default is 'rebeccapurple'.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' are accepted. The error is
        raised so that the plot does not accidentally gets rewritten.

    Returns
    -------
    None.
    """
    # parameter type check
    if isinstance(rewrite, bool) is not True:
        raise TypeError("'rewrite' should be boolean")
    plt.ioff()
    os.chdir(path)

    # Flatten the input list of pixels: if there are lists of pixels'
    # numbers inside the list, unpack them so that the output is a list
    # of numbers only
    pixels = _flatten(pixels)

    # Get the data files names for finding the appropriate '.feather'
    # file with timestamps differences, checking both options, depending
    # on which board was analyzed first
    folders = glob.glob("*#*")
    os.chdir(folders[0])
    # files_all = sorted(glob.glob("*.dat*"))
    files_all = glob.glob("*.dat*")
    # files_all.sort(key=lambda x: os.path.getmtime(x))
    files_all.sort(key=os.path.getmtime)

    feather_file_name1 = files_all[0][:-4] + "-"
    feather_file_name2 = "-" + files_all[-1][:-4]

    os.chdir("../{}".format(folders[1]))
    # files_all = sorted(glob.glob("*.dat*"))
    files_all = glob.glob("*.dat*")
    # files_all.sort(key=lambda x: os.path.getmtime(x))
    files_all.sort(key=os.path.getmtime)

    feather_file_name1 += files_all[-1][:-4]
    feather_file_name2 = files_all[0][:-4] + feather_file_name2
    os.chdir("..")

    # Check if plot exists and if it should be rewritten
    try:
        os.chdir(os.path.join(path, "results/delta_t"))
        if os.path.isfile(f"{feather_file_name1}_delta_t_grid.png"):
            if rewrite is True:
                print(
                    "\n! ! ! Plot of timestamp differences already"
                    "exists and will be rewritten ! ! !\n"
                )
            else:
                sys.exit(
                    "\nPlot already exists, 'rewrite' set to 'False', exiting."
                )

        elif os.path.isfile(f"{feather_file_name2}_delta_t_grid.png"):
            if rewrite is True:
                print(
                    "\n! ! ! Plot of timestamp differences already"
                    "exists and will be rewritten ! ! !\n"
                )
            else:
                sys.exit(
                    "\nPlot already exists, 'rewrite' set to 'False', exiting."
                )

        # os.chdir("../..")
    except FileNotFoundError:
        pass

    print(
        "\n> > > Plotting timestamps differences as a grid of histograms < < <"
    )

    plt.rcParams.update({"font.size": 27})
    # Prepare the grid for the plots based on the number of pixels
    # given
    if len(pixels) > 2:
        fig, axs = plt.subplots(
            len(pixels) - 1,
            len(pixels) - 1,
            figsize=(5.5 * len(pixels), 5.5 * len(pixels)),
        )
        for ax in axs:
            for x in ax:
                x.axes.set_axis_off()
    else:
        fig = plt.figure(figsize=(16, 16))

    # Check if the y limits of all plots should be the same
    if same_y is True:
        y_max_all = 0

    for q, _ in tqdm(enumerate(pixels), desc="Row in plot"):
        for w, _ in enumerate(pixels):
            if w <= q:
                continue
            if len(pixels) > 2:
                axs[q][w - 1].axes.set_axis_on()

            # Check if the file with timestamps differences is there
            feather_file_path1 = f"delta_ts_data/{feather_file_name1}.feather"
            feather_file_path2 = f"delta_ts_data/{feather_file_name2}.feather"
            feather_file, feather_file_name = (
                (feather_file_path1, feather_file_name1)
                if os.path.isfile(os.path.join(path, feather_file_path1))
                else (feather_file_path2, feather_file_name2)
            )
            if not os.path.isfile(os.path.join(path, feather_file)):
                raise FileNotFoundError(
                    "'.feather' file with timestamps differences was not found"
                )
            os.chdir(path)
            # Read data from Feather file
            try:
                data_to_plot = ft.read_feather(
                    os.path.join(
                        path, f"delta_ts_data/{feather_file_name}.feather"
                    ),
                    columns=[f"{pixels[q]},{pixels[w]}"],
                ).dropna()
            except ValueError:
                continue

            # Prepare the data for the plot
            data_to_plot = np.array(data_to_plot)
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot < range_left)
            )
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot > range_right)
            )

            # Bins should be in units of 17.857 ps - average bin width
            # of the LinoSPAD2 TDCs
            try:
                bins = np.arange(
                    np.min(data_to_plot),
                    np.max(data_to_plot),
                    2500 / 140 * multiplier,
                )
            except ValueError:
                print(
                    f"\nCouldn't calculate bins for {q}-{w} pair: probably "
                    "not enough delta ts."
                )
                continue

            if len(pixels) > 2:
                axs[q][w - 1].set_xlabel("\u0394t (ps)")
                axs[q][w - 1].set_ylabel("# of coincidences (-)")
                n, b, p = axs[q][w - 1].hist(
                    data_to_plot,
                    bins=bins,
                    color=color,
                )
            else:
                plt.xlabel("\u0394t (ps)")
                plt.ylabel("# of coincidences (-)")
                n, b, p = plt.hist(
                    data_to_plot,
                    bins=bins,
                    color=color,
                )

            # Find number of timestamps differences in a 2 ns window
            # around the peak (HBT or cross-talk)
            try:
                peak_max_pos = np.argmax(n).astype(np.intc)
                # 2 ns window around peak
                win = int(1000 / ((range_right - range_left) / 100))
                peak_max = np.sum(n[peak_max_pos - win : peak_max_pos + win])
            except ValueError:
                peak_max = 0

            if same_y is True:
                try:
                    y_max = np.max(n)
                except ValueError:
                    y_max = 0
                    print("\nCould not find maximum y value\n")
                if y_max_all < y_max:
                    y_max_all = y_max
                if len(pixels) > 2:
                    axs[q][w - 1].set_ylim(0, y_max + 4)
                else:
                    plt.ylim(0, y_max + 4)

            if len(pixels) > 2:
                axs[q][w - 1].set_xlim(range_left - 100, range_right + 100)
                axs[q][w - 1].set_title(
                    f"Pixels {pixels[q]},{pixels[w]}\nPeak in 2 ns "
                    f"window: {int(peak_max)}"
                )
            else:
                plt.xlim(range_left - 100, range_right + 100)

                plt.title(f"Pixels {pixels[0]},{256 + 255 - pixels[1]}")

            # Save the figure
            try:
                os.chdir("results/delta_t")
            except FileNotFoundError:
                os.makedirs("results/delta_t")
                os.chdir("results/delta_t")
            fig.tight_layout()  # for perfect spacing between the plots
            plt.savefig(
                "{name}_delta_t_grid.png".format(name=feather_file_name)
            )
            os.chdir("../..")

    print(
        "\n> > > Plot is saved as {file} in {path}< < <".format(
            file=feather_file_name + "_delta_t_grid.png",
            path=path + "/results/delta_t",
        )
    )



"""Module with scripts for sharing (and plotting) data in a compact form.

The module provides a simple and memory-friendly way of sharing data - 
that is processed '.dat' data in a form of a '.txt' file with number of
timestamps collected by each pixel and a '.feather' file with timestamp
differences. Additionally, function for plotting both the sensor
population plot and the delta t histograms are provided.

This file can also be imported as a module and contains the following
functions:

    * compact_share_feather - unpacks all '.dat' files in the given folder,
    collects the number of timestamps in each pixel and packs it into a
    '.txt' file, calculates timestamp differences and packs them into a 
    '.feather' file.

    * plot_shared - plots the sensor population plot from the '.txt'
    file.

    * collect_and_plot_timestamp_differences_shared_feather - plots the
    delta t histograms from the '.feather' file used for sharing data.

"""

import glob
import os
import sys
from math import ceil
from zipfile import ZipFile

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pyarrow import feather as ft
from tqdm import tqdm

from LinoSPAD2.functions import calc_diff as cd
from LinoSPAD2.functions import unpack as f_up
from LinoSPAD2.functions import utils


def compact_share_feather(
    path: str,
    pixels: list,
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int,
    delta_window: float = 50e3,
    include_offset: bool = True,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
):
    """Collect delta timestamp differences and sensor population, and
    save to a feather file.

    Unpacks data in the given folder, calculates timestamp differences
    and sensor population for the specified pixels, saving the timestamp
    differences to a '.feather' file and the sensor population to a '.txt'
    file. Both files are then zipped for compact output ready to share.

    Parameters
    ----------
    path : str
        Path to data files.
    pixels : list
        List of pixel numbers for which the timestamp differences should
        be calculated and saved, or list of two lists with pixel numbers
        for peak vs. peak calculations.
    rewrite : bool
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC calibration is
        applied. The default is True.
    absolute_timestamps : bool, optional
        Indicator for data with absolute timestamps. The default is
        False.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'firmware_version' are accepted. The first error is raised so
        that the files are not
        accidentally overwritten in the case of unclear input.

    Returns
    -------
    None.
    """

    # parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212b' or '2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError(
            "'daughterboard_number' should be string, either 'NL11' or 'A5'"
        )
    if isinstance(motherboard_number, str) is False:
        raise TypeError("'motherboard_number' should be string")

    os.chdir(path)

    files_all = sorted(glob.glob("*.dat*"))

    out_file_name = files_all[0][:-4] + "-" + files_all[-1][:-4]

    # check if feather file exists and if it should be rewrited
    feather_file = os.path.join(
        path, "compact_share", f"{out_file_name}.feather"
    )

    utils.file_rewrite_handling(feather_file, rewrite)

    # Collect the data for the required pixels
    print(
        "\n> > > Collecting data for delta t plot for the requested "
        "pixels and saving it to .feather in a cycle < < <\n"
    )
    if firmware_version == "2212s":
        # for transforming pixel number into TDC number + pixel
        # coordinates in that TDC
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    # Prepare array for sensor population
    valid_per_pixel = np.zeros(256, dtype=int)

    for i in tqdm(range(ceil(len(files_all))), desc="Collecting data"):
        file = files_all[i]

        # Prepare a dictionary for output
        deltas_all = {}

        # Unpack data for the requested pixels into dictionary
        if not absolute_timestamps:
            data_all = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            data_all, _ = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )

        # TODO use faster version
        deltas_all = cd.calculate_differences_2212(
            data_all, pixels, pix_coor, delta_window
        )

        # Collect sensor population
        for k in range(256):
            tdc, pix = np.argwhere(pix_coor == k)[0]
            valid_per_pixel[k] += np.count_nonzero(data_all[tdc][:, 0] == pix)

        # Save data as a .feather file in a cycle so data is not lost
        # in the case of failure close to the end
        data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
        del deltas_all
        data_for_plot_df = data_for_plot_df.T
        try:
            os.chdir("compact_share")
        except FileNotFoundError:
            os.mkdir("compact_share")
            os.chdir("compact_share")
        feather_file = f"{out_file_name}.feather"
        txt_file = (f"sen_pop_{out_file_name}.txt",)
        if os.path.isfile(feather_file):
            # Append delta t data
            existing_data = ft.read_feather(feather_file)
            combined_data = pd.concat(
                [existing_data, data_for_plot_df], axis=0
            )
            ft.write_feather(combined_data, feather_file)
            # Append sensor population data
            existing_data = np.genfromtxt(txt_file, delimiter="\t", dtype=int)
            combined_data = existing_data + valid_per_pixel
            np.savetxt(
                f"sen_pop_{out_file_name}.txt",
                combined_data,
                fmt="%d",
                delimiter="\t",
            )
        else:
            ft.write_feather(data_for_plot_df, feather_file)
            np.savetxt(
                f"sen_pop_{out_file_name}.txt",
                valid_per_pixel,
                fmt="%d",
                delimiter="\t",
            )
        os.chdir("..")

    os.chdir("compact_share")

    # Create a ZipFile Object
    with ZipFile(f"{out_file_name}.zip", "w") as zip_object:
        # Adding files that need to be zipped
        zip_object.write(f"{out_file_name}.feather")
        zip_object.write(f"sen_pop_{out_file_name}.txt")

        print(
            "\n> > > Timestamp differences are saved as {file}.feather and "
            "sensor population as sen_pop.txt in "
            "{path} < < <".format(
                file=out_file_name,
                path=path + "\delta_ts_data",
            )
        )


def plot_shared(
    path,
    daughterboard_number: str,
    motherboard_number: str,
    show_fig: bool = False,
    app_mask: bool = True,
    color: str = "rebeccapurple",
):
    """Plots sensor population from a '.txt' file.

    Plots the sensor population plot fro ma '.txt' file that is saved
    with the function above. Plot is saved in the
    'results/sensor_population' folder, which is created in the case it
    does not exist.

    Parameters
    ----------
    path : str
        Path to the '.txt' file with precompiled data.
    daughterboard_number : str
        The LinoSPAD2 daughterboard number.
    motherboard_number : str
        The LinoSPAD2 motherboard number.
    show_fig : bool, optional
        Switch for showing the plot. The default is False.
    app_mask : bool, optional
        Switch for applying the mask on warm/hot pixels. The default is
        True.
    color : str, optional
        Color for the plot. The default is 'salmon'.

    Raises
    ------
    IndexError
        _description_
    """
    os.chdir(path)

    try:
        file = glob.glob("*.txt*")[0]
    except IndexError:
        raise IndexError(".txt file not found - check the folder")

    file_name = file[:-4]

    data = np.genfromtxt(file, delimiter="")

    # Apply mask if requested
    if app_mask is True:
        path_to_back = os.getcwd()
        path_to_mask = os.path.realpath(__file__) + "/../.." + "/params/masks"
        os.chdir(path_to_mask)
        file_mask = glob.glob(
            "*{}_{}*".format(daughterboard_number, motherboard_number)
        )[0]
        mask = np.genfromtxt(file_mask).astype(int)
        data[mask] = 0
        os.chdir(path_to_back)

    if show_fig is True:
        plt.ion()
    else:
        plt.ioff()

    plt.rcParams.update({"font.size": 22})
    plt.figure(figsize=(16, 10))
    plt.xlabel("Pixel (-)")
    plt.ylabel("Photons (-)")
    plt.plot(data, "o-", color=color)

    try:
        os.chdir("results/sensor_population")
    except Exception:
        os.makedirs("results/sensor_population")
        os.chdir("results/sensor_population")
    plt.savefig("{}.png".format(file_name))
    os.chdir("..")


def collect_and_plot_timestamp_differences_shared_feather(
    path,
    pixels,
    rewrite: bool,
    range_left: int = -10e3,
    range_right: int = 10e3,
    step: int = 1,
    same_y: bool = False,
    color: str = "rebeccapurple",
):
    """Collect and plot timestamp differences from a '.feather' file.

    Plots timestamp differences from a '.feather' file as a grid of
    histograms and as a single plot. For plotting from the shared
    '.feather' files.

    Parameters
    ----------
    path : str
        Path to data files.
    pixels : list
        List of pixel numbers for which the timestamp differences
        should be plotted.
    rewrite : bool
        Switch for rewriting the plot if it already exists.
    range_left : int, optional
        Lower limit for timestamp differences, lower values are not used.
        The default is -10e3.
    range_right : int, optional
        Upper limit for timestamp differences, higher values are not used.
        The default is 10e3.
    step : int, optional
        Histogram binning multiplier. The default is 1.
    same_y : bool, optional
        Switch for plotting the histograms with the same y-axis.
        The default is False.
    color : str, optional
        Color for the plot. The default is 'rebeccapurple'.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' are accepted. The error is
        raised so that the plot does not accidentally gets rewritten.

    Returns
    -------
    None.
    """
    # parameter type check
    if isinstance(rewrite, bool) is not True:
        raise TypeError("'rewrite' should be boolean")
    plt.ioff()
    os.chdir(path)

    file = glob.glob("*.feather")[0]
    feather_file_name = file[:-8]

    print(
        "\n> > > Plotting timestamps differences as a grid of histograms < < <"
    )

    plt.rcParams.update({"font.size": 22})

    if len(pixels) > 2:
        fig, axs = plt.subplots(
            len(pixels) - 1,
            len(pixels) - 1,
            figsize=(5.5 * len(pixels), 5.5 * len(pixels)),
        )
        for ax in axs:
            for x in ax:
                x.axes.set_axis_off()
    else:
        fig = plt.figure(figsize=(14, 14))

    # check if the y limits of all plots should be the same
    if same_y is True:
        y_max_all = 0

    for q in tqdm(range(len(pixels)), desc="Row in plot"):
        for w in range(len(pixels)):
            if w <= q:
                continue
            if len(pixels) > 2:
                axs[q][w - 1].axes.set_axis_on()
            try:
                # keep only the required column in memory
                data_to_plot = ft.read_feather(
                    "{name}.feather".format(name=feather_file_name),
                    columns=["{},{}".format(pixels[q], pixels[w])],
                ).dropna()
            except ValueError:
                continue

            # prepare the data for plot
            data_to_plot = np.array(data_to_plot)
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot < range_left)
            )
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot > range_right)
            )

            try:
                bins = np.arange(
                    np.min(data_to_plot),
                    np.max(data_to_plot),
                    17.857 * step,
                )
            except ValueError:
                print(
                    "\nCouldn't calculate bins for {q}-{w} pair: probably not "
                    "enough delta ts.".format(q=q, w=w)
                )
                continue

            if len(pixels) > 2:
                axs[q][w - 1].set_xlabel("\u0394t (ps)")
                axs[q][w - 1].set_ylabel("# of coincidences (-)")
                n, b, p = axs[q][w - 1].hist(
                    data_to_plot,
                    bins=bins,
                    color=chosen_color,
                )
            else:
                plt.xlabel("\u0394t (ps)")
                plt.ylabel("# of coincidences (-)")
                n, b, p = plt.hist(
                    data_to_plot,
                    bins=bins,
                    color=color,
                )

            try:
                peak_max_pos = np.argmax(n).astype(np.intc)
                # 2 ns window around peak
                win = int(1000 / ((range_right - range_left) / 100))
                peak_max = np.sum(n[peak_max_pos - win : peak_max_pos + win])
            except ValueError:
                peak_max = None

            if same_y is True:
                try:
                    y_max = np.max(n)
                except ValueError:
                    y_max = 0
                    print("\nCould not find maximum y value\n")
                if y_max_all < y_max:
                    y_max_all = y_max
                if len(pixels) > 2:
                    axs[q][w - 1].set_ylim(0, y_max + 4)
                else:
                    plt.ylim(0, y_max + 4)

            if len(pixels) > 2:
                axs[q][w - 1].set_xlim(range_left - 100, range_right + 100)
                axs[q][w - 1].set_title(
                    "Pixels {p1},{p2}\nPeak in 2 ns window: {pp}".format(
                        p1=pixels[q], p2=pixels[w], pp=int(peak_max)
                    )
                )
            else:
                plt.xlim(range_left - 100, range_right + 100)
                plt.title(
                    "Pixels {p1},{p2}".format(p1=pixels[q], p2=pixels[w])
                )

            try:
                os.chdir("results/delta_t")
            except FileNotFoundError:
                os.makedirs("results/delta_t")
                os.chdir("results/delta_t")
            fig.tight_layout()  # for perfect spacing between the plots
            plt.savefig(
                "{name}_delta_t_grid.png".format(name=feather_file_name)
            )
            os.chdir("../..")
    print(
        "\n> > > Plot is saved as {file} in {path}< < <".format(
            file=feather_file_name + "_delta_t_grid.png",
            path=path + "/results/delta_t",
        )
    )



"""Module for working with the calibration data.

Functions in this module can be used to analyze the LinoSPAD2 data
to collect either a calibration matrix for compensating TDC 
nonlinearities or a calibration matrix for offset calibration. First
nonlinearity is introduced by nonequal bins of the 140-bin long TDC line
while the second - from the different-length electrical path in the PCB.

The calibration matrices (for TDC calibration) and arrays (for offset
calibration) should be put into ~/LinoSPAD2/src/LinoSPAD2/params
/calibration_data folder, where it will be pulled from by other
functions during analysis.

This file can also be imported as a module and contains the following
functions:

    * calibrate_and_save_TDC_calibration - calculate a calibration
    matrix of the TDC calibrations and save it as a '.csv' table.

    * unpack_data_for_offset_calibration - unpack binary data applying
    TDC calibration in the process. Used for calculations of the offset
    calibration.

    * save_offset_timestamp_differences - calculate and save timestamps
    differences for pairs of pixels 0, 4 to 255, and 1 to 3 for finding
    the delta t peaks for calculating the offset calibration.

    * calculate_and_save_offset_calibration - calculate and save the 256
    offset compensations for all pixels of the given LinoSPAD2 sensor
    half. The output is saved as a .npy file for later use.

    * load_calibration_data - load the calibration matrix from a '.csv'
    table.

"""

import glob
import os
import sys
import time
from math import ceil

import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from tqdm import tqdm


def calibrate_and_save_TDC_data(
    path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 1000,
):
    """Calculate and save calibration data for TDC nonlinearities.

    Function for calculating the calibration matrix and saving it into a
    '.csv' file. The data files used for the calculation should be taken
    with the sensor uniformly illuminated by ambient light.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Versions "2208", "2212s" (skip), and
        "2212b" (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The
        default is 1000.

    Returns
    -------
    None.

    Raises
    ------
    TypeError
        If 'daughterboard_number', 'motherboard_number', or
        'firmware_version' parameters are not of string type.

    Notes
    -----
    The resulting calibration matrix is saved as a '.csv' file, with the
    filename formatted as "TDC_{db}_{mb}_{fw_ver}.csv", where {db},
    {mb}, and {fw_ver} represent the daughterboard number, motherboard
    number, and firmware version, respectively.
    """

    # Parameter type check
    if not isinstance(daughterboard_number, str):
        raise TypeError("'daughterboard_number' should be a string.")
    if not isinstance(motherboard_number, str):
        raise TypeError("'motherboard_number' should be a string.")
    if not isinstance(firmware_version, str):
        raise TypeError(
            "'firmware_version' should be a string, '2208', '2212b' or"
            "'2212s'."
        )

    os.chdir(path)
    files = glob.glob("*.dat")

    # Go over all '.dat' files
    for j, file in enumerate(
        tqdm(
            files,
            desc="Calculating TDC calibration, going through files",
        )
    ):
        if firmware_version == "2208":
            # read data by 32 bit words
            raw_data = np.fromfile(file, dtype=np.uint32)
            # lowest 28 bits are the timestamp; convert to ps
            data = (raw_data & 0xFFFFFFF).astype(int) % 140
            # mask nonvalid data with '-1'; 0x80000000 - the 31st, validity bin
            data[np.where(raw_data < 0x80000000)] = -1
            # number of acquisition cycles
            cycles = int(len(data) / timestamps / 256)

            data_matrix = (
                data.reshape(cycles, 256, timestamps)
                .transpose((1, 0, 2))
                .reshape(256, timestamps * cycles)
            )

            # calibration matrix
            cal_mat = np.zeros((256, 140))
            bins = np.arange(0, 141, 1)

            for i in range(256):
                # sort the data into 140 bins
                counts, _ = np.histogram(data_matrix[i], bins=bins)
                # redefine the bin edges using the bin population from above
                cal_mat[i] = np.cumsum(counts) / np.cumsum(counts).max() * 2500

            cal_mat_df = pd.DataFrame(cal_mat)
            cal_mat_df.to_csv(
                f"TDC_{daughterboard_number}_{motherboard_number}_"
                f"{firmware_version}_{j}.csv"
            )

        elif firmware_version == "2212b" or firmware_version == "2212s":
            # read data by 32 bit words
            raw_data = np.fromfile(file, dtype=np.uint32)
            # lowest 28 bits are the timestamp; convert to ps
            data_t = (raw_data & 0xFFFFFFF).astype(int) % 140
            # pix address in the given TDC is 2 bits above timestamp
            data_p = ((raw_data >> 28) & 0x3).astype(np.longlong)
            data_t[np.where(raw_data < 0x80000000)] = -1
            # number of acquisition cycle in each data file
            cycles = int(len(data_t) / timestamps / 65)
            # transform into matrix 65 by cycles*timestamps
            data_matrix_p = (
                data_p.reshape(cycles, 65, timestamps)
                .transpose((1, 0, 2))
                .reshape(65, timestamps * cycles)
            )

            data_matrix_t = (
                data_t.reshape(cycles, 65, timestamps)
                .transpose((1, 0, 2))
                .reshape(65, timestamps * cycles)
            )

            # cut the 65th TDC that does not hold any actual data from pixels
            data_matrix_p = data_matrix_p[:-1]
            data_matrix_t = data_matrix_t[:-1]

            data_all = np.stack((data_matrix_p, data_matrix_t), axis=2).astype(
                np.longlong
            )

            # calibration matrix
            cal_mat = np.zeros((256, 140))
            bins = np.arange(0, 141, 1)

            if firmware_version == "2212b":
                pix_coor = np.arange(256).reshape(64, 4)
            else:
                pix_coor = np.arange(256).reshape(4, 64).T

            for i in range(256):
                # transform pixel number to TDC number and pixel coordinates in
                # that TDC (from 0 to 3)
                tdc, pix = np.argwhere(pix_coor == i)[0]
                # find data from that pixel
                ind = np.where(data_all[tdc].T[0] == pix)[0]
                # cut non-valid timestamps ('-1's)
                ind = ind[np.where(data_all[tdc].T[1][ind] >= 0)[0]]
                if not np.any(ind):
                    continue

                counts, _ = np.histogram(data_all[tdc].T[1][ind], bins=bins)
                cal_mat[i] = np.cumsum(counts) / np.cumsum(counts).max() * 2500

            cal_mat_df = pd.DataFrame(cal_mat)
            cal_mat_df.to_csv(
                f"TDC_{daughterboard_number}_{motherboard_number}_"
                f"{firmware_version}_{j}.csv"
            )

    # Combine all '.csv' files and average
    files_csv = glob.glob("*.csv")

    data_csv = np.zeros((256, 140))

    for i, file_csv in enumerate(files_csv):
        data_csv += pd.read_csv(file_csv, index_col=0)

    data_csv = data_csv / (i + 1)

    # Save the averaged matrix of calibration data into a '.csv' file
    data_csv.to_csv(
        f"TDC_{daughterboard_number}_{motherboard_number}_"
        f"{firmware_version}.csv"
    )

    # Remove the numbered '.csv' files
    file_pattern = f"TDC_{daughterboard_number}_{motherboard_number}"
    f"_{firmware_version}_*.csv"
    files_to_delete = glob.glob(file_pattern)
    for file_to_delete in files_to_delete:
        os.remove(file_to_delete)


# TODO: probably not required - remove
# def calibrate_and_save_TDC_data_DNL_corrected(
#     path: str,
#     daughterboard_number: str,
#     motherboard_number: str,
#     firmware_version: str,
#     timestamps: int = 1000,
# ):
#     """Calculate and save calibration data for TDC nonlinearities.

#     Function for calculating the calibration matrix and saving it into a
#     '.csv' file. The data files used for the calculation should be taken
#     with the sensor uniformly illuminated by ambient light.

#     Parameters
#     ----------
#     path : str
#         Path to the folder with the '.dat' data files.
#     daughterboard_number : str
#         LinoSPAD2 daughterboard number.
#     motherboard_number : str
#         LinoSPAD2 motherboard number, including the "#".
#     firmware_version : str
#         LinoSPAD2 firmware version. Versions "2208", "2212s" (skip), and
#         "2212b" (block) are recognized.
#     timestamps : int, optional
#         Number of timestamps per acquisition cycle per pixel. The
#         default is 1000.

#     Returns
#     -------
#     None.

#     Raises
#     ------
#     TypeError
#         If 'daughterboard_number', 'motherboard_number', or
#         'firmware_version' parameters are not of string type.

#     Notes
#     -----
#     The resulting calibration matrix is saved as a '.csv' file, with the
#     filename formatted as "TDC_{db}_{mb}_{fw_ver}.csv", where {db},
#     {mb}, and {fw_ver} represent the daughterboard number, motherboard
#     number, and firmware version, respectively.
#     """

#     # Parameter type check
#     if not isinstance(daughterboard_number, str):
#         raise TypeError("'daughterboard_number' should be a string.")
#     if not isinstance(motherboard_number, str):
#         raise TypeError("'motherboard_number' should be a string.")
#     if not isinstance(firmware_version, str):
#         raise TypeError(
#             "'firmware_version' should be a string, '2208', '2212b' or"
#             "'2212s'."
#         )

#     os.chdir(path)
#     files = glob.glob("*.dat")

#     # Go over all '.dat' files
#     for j, file in enumerate(
#         tqdm(
#             files,
#             desc="Calculating TDC calibration, going through files",
#         )
#     ):
#         if firmware_version == "2208":
#             # read data by 32 bit words
#             raw_data = np.fromfile(file, dtype=np.uint32)
#             # lowest 28 bits are the timestamp; convert to ps
#             data = (raw_data & 0xFFFFFFF).astype(int) % 140
#             # mask nonvalid data with '-1'; 0x80000000 - the 31st, validity bin
#             data[np.where(raw_data < 0x80000000)] = -1
#             # number of acquisition cycles
#             cycles = int(len(data) / timestamps / 256)

#             data_matrix = (
#                 data.reshape(cycles, 256, timestamps)
#                 .transpose((1, 0, 2))
#                 .reshape(256, timestamps * cycles)
#             )

#             # calibration matrix
#             cal_mat = np.zeros((256, 140))
#             bins = np.arange(0, 141, 1)

#             for i in range(256):
#                 # sort the data into 140 bins
#                 counts, _ = np.histogram(data_matrix[i], bins=bins)
#                 # redefine the bin edges using the bin population from above
#                 cal_mat[i] = np.cumsum(counts) / np.cumsum(counts).max() * 2500

#             cal_mat_df = pd.DataFrame(cal_mat)
#             cal_mat_df.to_csv(
#                 f"TDC_{daughterboard_number}_{motherboard_number}_"
#                 f"{firmware_version}_{j}.csv"
#             )

#         elif firmware_version == "2212b" or firmware_version == "2212s":
#             # read data by 32 bit words
#             raw_data = np.fromfile(file, dtype=np.uint32)
#             # lowest 28 bits are the timestamp; convert to ps
#             data_t = (raw_data & 0xFFFFFFF).astype(int) % 140
#             # pix address in the given TDC is 2 bits above timestamp
#             data_p = ((raw_data >> 28) & 0x3).astype(np.longlong)
#             data_t[np.where(raw_data < 0x80000000)] = -1
#             # number of acquisition cycle in each data file
#             cycles = int(len(data_t) / timestamps / 65)
#             # transform into matrix 65 by cycles*timestamps
#             data_matrix_p = (
#                 data_p.reshape(cycles, 65, timestamps)
#                 .transpose((1, 0, 2))
#                 .reshape(65, timestamps * cycles)
#             )

#             data_matrix_t = (
#                 data_t.reshape(cycles, 65, timestamps)
#                 .transpose((1, 0, 2))
#                 .reshape(65, timestamps * cycles)
#             )

#             # cut the 65th TDC that does not hold any actual data from pixels
#             data_matrix_p = data_matrix_p[:-1]
#             data_matrix_t = data_matrix_t[:-1]

#             data_all = np.stack((data_matrix_p, data_matrix_t), axis=2).astype(
#                 np.longlong
#             )

#             # calibration matrix
#             cal_mat = np.zeros((256, 140))
#             bins = np.arange(0, 141, 1)

#             if firmware_version == "2212b":
#                 pix_coor = np.arange(256).reshape(64, 4)
#             else:
#                 pix_coor = np.arange(256).reshape(4, 64).T

#             for i in range(256):
#                 # transform pixel number to TDC number and pixel coordinates in
#                 # that TDC (from 0 to 3)
#                 tdc, pix = np.argwhere(pix_coor == i)[0]
#                 # find data from that pixel
#                 ind = np.where(data_all[tdc].T[0] == pix)[0]
#                 # cut non-valid timestamps ('-1's)
#                 ind = ind[np.where(data_all[tdc].T[1][ind] >= 0)[0]]
#                 if not np.any(ind):
#                     continue

#                 counts, _ = np.histogram(data_all[tdc].T[1][ind], bins=bins)

#                 # TODO
#                 file_DNL = os.path.join(
#                     r"C:\Users\bruce\Documents\GitHub\LinoSPAD2\src\LinoSPAD2\params\calibration_data",
#                     f"{daughterboard_number}_{motherboard_number}_{firmware_version}_DNL.npy",
#                 )
#                 M = np.load(file_DNL)
#                 counts_calib = M[tdc] @ counts
#                 cal_mat[i] = (
#                     np.cumsum(counts_calib)
#                     / np.cumsum(counts_calib).max()
#                     * 2500
#                 )

#             cal_mat_df = pd.DataFrame(cal_mat)
#             cal_mat_df.to_csv(
#                 f"TDC_DNL_{daughterboard_number}_{motherboard_number}_"
#                 f"{firmware_version}_{j}.csv"
#             )

#     # Combine all '.csv' files and average
#     files_csv = glob.glob("*.csv")

#     data_csv = np.zeros((256, 140))

#     for i, file_csv in enumerate(files_csv):
#         data_csv += pd.read_csv(file_csv, index_col=0)

#     data_csv = data_csv / (i + 1)

#     # Save the averaged matrix of calibration data into a '.csv' file
#     data_csv.to_csv(
#         f"TDC_DNL_{daughterboard_number}_{motherboard_number}_"
#         f"{firmware_version}.csv"
#     )

#     # Remove the numbered '.csv' files
#     file_pattern = (
#         f"TDC_DNL_{daughterboard_number}_"
#         f"{motherboard_number}_{firmware_version}_*.csv"
#     )
#     files_to_delete = glob.glob(file_pattern)
#     for file_to_delete in files_to_delete:
#         os.remove(file_to_delete)


def unpack_data_for_offset_calibration(
    path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 1000,
):
    """Unpack data for offset calibration from LinoSPAD2 firmware version 2212.

    Unpacks binary-encoded data from LinoSPAD2 firmware version 2212.
    Returns a 3D array where rows are TDC numbers, columns are the data,
    and each cell contains a pixel number in the TDC (from 0 to 3) and the
    timestamp recorded by that pixel. TDC calibration is applied as it
    is necessary for the offset calibration to succeed.

    Parameters
    ----------
    path : str
        Path to the '.dat' data file.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Either '2212s' or '2212b' are accepted.
    timestamps : int, optional
        Number of timestamps per TDC per acquisition cycle. The default
        is 1000.

    Raises
    ------
    TypeError
        If 'daughterboard_number', 'motherboard_number', or
        'firmware_version' parameters are not of string type.
    FileNotFoundError
        If no calibration data is found, raise an error.

    Returns
    -------
    data_all : array-like
        3D array of pixel coordinates in the TDC and the timestamps.

    Notes
    -----
    This function unpacks binary-encoded data from LinoSPAD2 firmware
    version 2212 for offset calibration. The resulting data is returned
    as a 3D array where rows are TDC numbers, columns are the data, and
    each cell contains a pixel number in the TDC (from 0 to 3) and the
    timestamp recorded by that pixel. TDC calibration is applied as it
    is necessary for the offset calibration to succeed. The calibration
    data is loaded from the specified path based on the daughterboard,
    motherboard, and firmware version parameters.
    """
    # Parameter type check
    if not isinstance(daughterboard_number, str):
        raise TypeError("'daughterboard_number' should be a string.")
    if not isinstance(motherboard_number, str):
        raise TypeError("'motherboard_number' should be a string.")
    if not isinstance(firmware_version, str):
        raise TypeError(
            "'firmware_version' should be a string, '2212s' or '2212b'."
        )

    # Unpack binary data
    raw_file_data = np.fromfile(path, dtype=np.uint32)
    # Timestamps are lower 28 bits
    data_timestamps = (raw_file_data & 0xFFFFFFF).astype(np.longlong)
    # Pix address in the given TDC is 2 bits above timestamp
    data_pix = ((raw_file_data >> 28) & 0x3).astype(np.longlong)
    data_timestamps[np.where(raw_file_data < 0x80000000)] = -1
    # Number of acquisition cycle in each data file
    cycles = int(len(data_timestamps) / timestamps / 65)
    # Transform into matrix 65 by cycles*timestamps
    data_matrix_pix = (
        data_pix.reshape(cycles, 65, timestamps)
        .transpose((1, 0, 2))
        .reshape(65, timestamps * cycles)
    )

    data_matrix_timestamps = (
        data_timestamps.reshape(cycles, 65, timestamps)
        .transpose((1, 0, 2))
        .reshape(65, timestamps * cycles)
    )
    # Cut the 65th TDC that does not hold any actual data from pixels
    data_matrix_pix = data_matrix_pix[:-1]
    data_matrix_timestamps = data_matrix_timestamps[:-1]
    # Insert '-2' at the end of each cycle
    data_matrix_pix = np.insert(
        data_matrix_pix,
        np.linspace(timestamps, cycles * timestamps, cycles).astype(
            np.longlong
        ),
        -2,
        1,
    )

    data_matrix_timestamps = np.insert(
        data_matrix_timestamps,
        np.linspace(timestamps, cycles * timestamps, cycles).astype(
            np.longlong
        ),
        -2,
        1,
    )
    # Combine both matrices into a single one, where each cell holds pix
    # coordinates in the TDC and the timestamp
    data_all = np.stack(
        (data_matrix_pix, data_matrix_timestamps), axis=2
    ).astype(np.longlong)

    # Path to the calibration data directory
    pix_coordinate_array = np.arange(256).reshape(64, 4)
    path_calibration_data = (
        os.path.realpath(__file__) + "/../.." + "/params/calibration_data"
    )

    try:
        cal_matrix = load_calibration_data(
            path_calibration_data,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            include_offset=False,
        )
    except FileNotFoundError:
        raise FileNotFoundError(
            "No .csv file with the calibration data was found, "
            "check the path or run the calibration."
        )

    for i in range(256):
        # Transform pixel number to TDC number and pixel coordinates in
        # that TDC (from 0 to 3)
        tdc, pix = np.argwhere(pix_coordinate_array == i)[0]
        # Find data from that pixel
        ind = np.where(data_all[tdc].T[0] == pix)[0]
        # Cut non-valid timestamps ('-1's)
        ind = ind[np.where(data_all[tdc].T[1][ind] >= 0)[0]]
        if not np.any(ind):
            continue

        data_all[tdc].T[1][ind] = (
            data_all[tdc].T[1][ind] - data_all[tdc].T[1][ind] % 140
        ) * 2500 / 140 + cal_matrix[i, (data_all[tdc].T[1][ind] % 140)]

    return data_all


def save_offset_timestamp_differences(
    path: str,
    pixels: list,
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 1000,
    delta_window: float = 50e3,
):
    """Calculate and save timestamp differences into '.csv' file.

    Unpacks data, calculates timestamp differences for the requested
    pixels, and saves them into a '.csv' table. Works with firmware
    version 2212. Calculates delta ts with TDC calibration applied
    for calculations of offset compensations.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    pixels : list
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    rewrite : bool
        Switch for overwriting the '.csv' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 1000.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'daughterboard_number', 'motherboard_number', and
        'firmware_version' are accepted. The first error is raised so
        that the plot does not accidentally get rewritten in the case
        no clear input was given.

    Returns
    -------
    None.
    """
    # Parameter type check
    if not isinstance(pixels, list):
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists."
        )
    if not isinstance(firmware_version, str):
        raise TypeError(
            "'firmware_version' should be a string, '2212b' or '2208'."
        )
    if not isinstance(rewrite, bool):
        raise TypeError("'rewrite' should be boolean.")
    if not isinstance(daughterboard_number, str):
        raise TypeError(
            "'daughterboard_number' should be a string, either 'NL11' or 'A5'."
        )

    os.chdir(path)

    all_files = glob.glob("*.dat*")

    output_file_name = all_files[0][:-4] + "-" + all_files[-1][:-4]

    # Check if csv file exists and if it should be rewritten
    try:
        os.chdir("offset_deltas")
        if os.path.isfile(f"{output_file_name}.csv"):
            if rewrite:
                print(
                    "\n! ! ! CSV file with timestamps differences already "
                    "exists and will be overwritten ! ! !\n"
                )
                for i in range(5):
                    print(f"\n! ! ! Deleting the file in {5 - i} ! ! !\n")
                    time.sleep(1)
                os.remove(f"{output_file_name}.csv")
            else:
                sys.exit(
                    "\n CSV file already exists, 'rewrite' set to"
                    " 'False', exiting."
                )
        os.chdir("..")
    except FileNotFoundError:
        pass

    # Collect the data for the required pixels
    print(
        "\n> > > Collecting data for delta t plot for the requested "
        "pixels and saving it to .csv in a cycle < < <\n"
    )
    if firmware_version == "2212s":
        pix_coordinates = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coordinates = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    # Mask the hot/warm pixels
    # mask = utils.apply_mask(daughterboard_number, motherboard_number)

    # Check if 'pixels' is one or two peaks, swap their positions if
    # needed
    if isinstance(pixels[0], list):
        pixels_left = pixels[0]
        pixels_right = pixels[1]
        # Check if pixels from the first list are to the left of the right
        # (peaks are not mixed up)
        if pixels_left[-1] > pixels_right[0]:
            pixels_left, pixels_right = pixels_right, pixels_left
    elif isinstance(pixels[0], int):
        pixels_left = pixels
        pixels_right = pixels

    for i in tqdm(range(ceil(len(all_files))), desc="Collecting data"):
        file = all_files[i]

        # Prepare a dictionary for output
        deltas_all = {}

        # Unpack data for the requested pixels into dictionary
        data_all = unpack_data_for_offset_calibration(
            file,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
        )

        # Calculate and collect timestamp differences
        for q in pixels_left:
            for w in pixels_right:
                if w <= q:
                    continue
                deltas_all[f"{q},{w}"] = []
                # Find end of cycles
                cycler = np.argwhere(data_all[0].T[0] == -2)
                cycler = np.insert(cycler, 0, 0)
                # First pixel in the pair
                tdc1, pix_c1 = np.argwhere(pix_coordinates == q)[0]
                pix1 = np.where(data_all[tdc1].T[0] == pix_c1)[0]
                # Second pixel in the pair
                tdc2, pix_c2 = np.argwhere(pix_coordinates == w)[0]
                pix2 = np.where(data_all[tdc2].T[0] == pix_c2)[0]
                # Get timestamp for both pixels in the given cycle
                for cyc in range(len(cycler) - 1):
                    pix1_ = pix1[
                        np.logical_and(
                            pix1 > cycler[cyc], pix1 < cycler[cyc + 1]
                        )
                    ]
                    if not np.any(pix1_):
                        continue
                    pix2_ = pix2[
                        np.logical_and(
                            pix2 > cycler[cyc], pix2 < cycler[cyc + 1]
                        )
                    ]
                    if not np.any(pix2_):
                        continue
                    # Calculate delta t
                    tmsp1 = data_all[tdc1].T[1][
                        pix1_[np.where(data_all[tdc1].T[1][pix1_] > 0)[0]]
                    ]
                    tmsp2 = data_all[tdc2].T[1][
                        pix2_[np.where(data_all[tdc2].T[1][pix2_] > 0)[0]]
                    ]
                    for t1 in tmsp1:
                        deltas = tmsp2 - t1
                        ind = np.where(np.abs(deltas) < delta_window)[0]
                        deltas_all[f"{q},{w}"].extend(deltas[ind])

        # Save data as a .csv file in a cycle so data is not lost
        # in the case of a failure close to the end
        data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
        del deltas_all
        data_for_plot_df = data_for_plot_df.T
        try:
            os.chdir("offset_deltas")
        except FileNotFoundError:
            os.mkdir("offset_deltas")
            os.chdir("offset_deltas")
        csv_file = glob.glob(f"*{output_file_name}.csv*")
        if csv_file:
            data_for_plot_df.to_csv(
                f"Offset_{output_file_name}.csv",
                mode="a",
                index=False,
                header=False,
            )
        else:
            data_for_plot_df.to_csv(
                f"Offset_{output_file_name}.csv", index=False
            )
        os.chdir("..")

    if os.path.isfile(path + f"/offset_deltas/{output_file_name}.csv") is True:
        print(
            "\n> > > Timestamp differences are saved as"
            f"{output_file_name}.csv in "
            f"{os.path.join(path, 'offset_deltas')} < < <"
        )
    else:
        print("File wasn't generated. Check input parameters.")


def calculate_and_save_offset_calibration(
    path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 1000,
):
    """Calculate offset calibration and save as .npy.

    Calculate offset calibration for all 256 pixels for the given
    motherboard-daughterboard and firmware version.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version.
    timestamps : int, optional
        Number of timestamps per cycle per TDC. The default 1000.
    """

    def gauss(x, A, x0, sigma):
        return A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))

    # Calculate delta ts for pixels 0 and 4-255
    save_offset_timestamp_differences(
        path,
        pixels=[[0], [x for x in range(1, 256)]],
        rewrite=True,
        daughterboard_number=daughterboard_number,
        motherboard_number=motherboard_number,
        firmware_version=firmware_version,
        timestamps=timestamps,
    )
    os.chdir(path + r"/offset_deltas/")
    file_csv = glob.glob("*Offset_*.csv*")[0]
    dt_all = np.array(pd.read_csv(file_csv))
    os.chdir("..")

    peak_positions_3_256 = np.zeros(256)
    peak_positions_1_4 = np.zeros(256)

    # Fit to find where the peak ends up.
    for i in range(255):
        dt_nonan_arr = dt_all[:, i][~np.isnan(dt_all[:, i])]
        if dt_nonan_arr.size == 0:
            continue
        else:
            bins = np.arange(
                np.min(dt_nonan_arr), np.max(dt_nonan_arr), 2500 / 140
            )

            counts, bin_edges = np.histogram(dt_nonan_arr, bins=bins)
            bin_centers = 0.5 * (bin_edges[1:] + bin_edges[:-1])

            n_max = np.argmax(counts)
            arg_max = (bin_edges[n_max] + bin_edges[n_max + 1]) / 2
            sigma = 200

            parameters, covariance = curve_fit(
                gauss, bin_centers, counts, p0=[max(counts), arg_max, sigma]
            )

            peak_positions_3_256[i] = parameters[1]
            peak_positions_3_256[0:3] = 0

    # Calculate delta ts for pixels 1,2,3
    save_offset_timestamp_differences(
        path,
        pixels=[[1, 2, 3], [4]],
        rewrite=True,
        daughterboard_number=daughterboard_number,
        motherboard_number=motherboard_number,
        firmware_version=firmware_version,
        timestamps=timestamps,
    )
    os.chdir(path + r"/offset_deltas/")
    file_csv = glob.glob("*.csv*")[0]
    dt_all = np.array(pd.read_csv(file_csv))
    os.chdir("..")

    peak_positions_1_4 = np.zeros(256)

    # Fit to find where the peak ends up.
    for i in range(3):
        dt_nonan_arr = dt_all[:, i][~np.isnan(dt_all[:, i])]
        if dt_nonan_arr.size == 0:
            continue
        else:
            bins = np.arange(
                np.min(dt_nonan_arr), np.max(dt_nonan_arr), 2500 / 140
            )

            counts, bin_edges = np.histogram(dt_nonan_arr, bins=bins)
            bin_centers = 0.5 * (bin_edges[1:] + bin_edges[:-1])

            n_max = np.argmax(counts)
            arg_max = (bin_edges[n_max] + bin_edges[n_max + 1]) / 2
            sigma = 200

            parameters, covariance = curve_fit(
                gauss, bin_centers, counts, p0=[max(counts), arg_max, sigma]
            )

            peak_positions_1_4[i] = parameters[1]

    peak_positions = peak_positions_1_4 + peak_positions_3_256
    print(peak_positions)
    print(peak_positions)

    # Indices for a system of linear equations for offset calculation.
    # Last equation is for setting the average offset equal to zero.
    a = np.zeros((256, 256))
    for i in range(3, 255):
        a[i][0] = 1
        a[i][i + 1] = -1
    a[0][1] = 1
    a[1][2] = 1
    a[2][3] = 1
    a[0][4] = -1
    a[1][4] = -1
    a[2][4] = -1
    a[-1] = 1

    # Solving the system of equations, the result are offsets
    offsets = np.linalg.solve(a, peak_positions)

    np.save(
        f"Offset_{daughterboard_number}_{motherboard_number}"
        f"_{firmware_version}.npy",
        offsets,
    )


def load_calibration_data(
    calibration_path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    include_offset: bool = False,
):
    """Load the calibration data.

    Parameters
    ----------
    calibration_path : str
        Path to the '.csv' file with the calibration matrix.
    daughterboard_number: str
        The LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version: str
        LinoSPAD2 firmware version.
    include_offset : bool, optional
        Switch for including the offset calibration. The default is
        True.

    Returns
    -------
    data_matrix : numpy.ndarray
        256x140 matrix containing the calibrated data.
    offset_arr : numpy.ndarray, optional
        Array of 256 offset values, one for each pixel. Returned only if
        include_offset is True.
    """

    path_to_backup = os.getcwd()
    os.chdir(calibration_path)

    # Compensating for TDC nonlinearities
    try:
        file_TDC = glob.glob(
            f"*TDC_{daughterboard_number}_{motherboard_number}"
            f"_{firmware_version}*"
        )[0]
    except IndexError as exc:
        raise FileNotFoundError(
            f"TDC calibration for {daughterboard_number}, "
            f"{motherboard_number}, and {firmware_version} is not found"
        ) from exc

    # Compensating for offset
    if include_offset:
        try:
            file_offset = glob.glob(
                f"*Offset_{daughterboard_number}_{motherboard_number}"
                f"_{firmware_version}*"
            )[0]
        except IndexError:
            raise FileNotFoundError(
                "No .npy file with offset calibration data was found"
            )
        offset_arr = np.load(file_offset)

    # Skipping the first row of TDC bins' numbers
    data_matrix_TDC = np.genfromtxt(file_TDC, delimiter=",", skip_header=1)

    # Cut the first column which is pixel numbers
    data_matrix_TDC = np.delete(data_matrix_TDC, 0, axis=1)

    os.chdir(path_to_backup)

    return (data_matrix_TDC, offset_arr) if include_offset else data_matrix_TDC



"""Module of utility functions.

This module provides a set of functions used in other LinoSPAD2 functions,
covering tasks such as applying masks, unpickling and displaying
matplotlib figures, defining Gaussian functions, fitting them to data,
and transforming lists of pixels.

This file can also be imported as a module and contains the following
functions:

    * apply_mask - Apply a mask to the given data based on the
    daughterboard and motherboard numbers.

    * unpickle_plot - Unpickle and display a matplotlib figure based on
    the specified type.

    * gaussian - Gaussian function for curve fitting.

    * fit_gaussian - Fit Gaussian function to data and return optimal
    parameters and covariance.

    * pixel_list_transform - Transform a list of pixels into two separate
    lists based on input type.
    
    * correct_pixels_address - correct the pixel addressing, the output
    has the same dimensions as the input. Should be used for motherboard
    on side "23" of the daughterboard.

    * file_rewrite_handling - based on the file name given and the
    boolean paramter 'rewrite', handles the file overwriting based on
    its' existence. Introduced for clearer code and modularity.
    
    * error_propagation_division - propagate error for the division
    operation
    
    * combine_feather_files - Combine multiple '.feather' files into one.
    Can be used in delta_t functions for collecting timestamp differences
    to alleviate bottlenecks in saving results or for combining files
    manually for testing. 
    

"""

import glob
import os
import pickle
import sys
import time
from typing import List

import matplotlib
import numpy as np
import pandas as pd
from pyarrow import feather as ft
from scipy.optimize import curve_fit


def apply_mask(
    daughterboard_number: str, motherboard_number: str
) -> np.ndarray:
    """Find and return mask for the requested motherboard.

    Parameters
    ----------
    daughterboard_number : str
        The LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".

    Returns
    -------
    mask : np.ndarray
        The mask array generated from the given daughterboard and motherboard numbers.
    """

    path_to_back = os.getcwd()
    path_to_mask = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "..",
        "params",
        "masks",
    )
    os.chdir(path_to_mask)
    file_mask = glob.glob(f"*{daughterboard_number}_{motherboard_number}*")[0]
    mask = np.genfromtxt(file_mask).astype(int)
    os.chdir(path_to_back)

    return mask


def unpickle_plot(path: str, plot_type: str, interactive: bool = True) -> None:
    """Unpickle and display a matplotlib figure.

    This function loads a previously pickled matplotlib figure from the
    specified path and displays it. Optionally, it can switch to an
    interactive backend.

    Parameters
    ----------
    path : str
        Path to the folder containing the pickled matplotlib figure.
    plot_type : str
        Type of the plot. Should be one of 'sensor_population', 'fits',
        or 'delta_t'.
    interactive : bool, optional
        Switch for making the plot interactive by using QtAgg matplotlib
        backend. The default is True.

    Raises
    ------
    TypeError
        Raised if the 'type' parameter is not one of 'sensor_population',
        'fits', or 'delta_t'.

    Returns
    -------
    None.

    """
    if plot_type not in ["sensor_population", "fits", "delta_t"]:
        raise TypeError(
            "Type should be 'sensor_population', 'fits', or 'delta_t'"
        )

    os.chdir(path)
    files = glob.glob("*.dat*")
    plot_name = files[0][:-4] + "-" + files[-1][:-4]

    os.chdir(f"results/{plot_type}")

    if interactive:
        matplotlib.use("QtAgg")

    with open(f"{plot_name}.pickle", "rb") as file:
        fig = pickle.load(file)
        fig.show()


def gaussian(x, amp, mu, sigma, bkg):
    """Gaussian function.

    Parameters
    ----------
    x : array-like
        The input data.
    amp : float
        Amplitude of the Gaussian.
    mu : float
        Mean (center) of the Gaussian.
    sigma : float
        Standard deviation of the Gaussian.
    bkg : float
        Background offset.

    Returns
    -------
    array-like
        The computed Gaussian values.

    """
    return amp * np.exp(-((x - mu) ** 2) / (2 * sigma**2)) + bkg


def fit_gaussian(x, y):
    """Fit Gaussian function to data.

    Parameters
    ----------
    x : array-like
        The x-axis data.
    y : array-like
        The y-axis data.

    Returns
    -------
    tuple
        A tuple containing two elements:
        - popt : array-like
            Optimal values for the parameters so that the sum of the
            squared residuals of `gaussian(x, *popt) - y` is minimized.
        - pcov : 2D array-like
            Estimated covariance of `popt`. The diagonal elements represent
            the variance of the parameter estimates.

    """
    # Initial guess for the parameters
    amp_guess = np.max(y)
    mu_guess = x[np.argmax(y)]
    # As std sometimes gives nonsense, 150 ps is added as balancing
    sigma_guess = min(np.std(x), 150)
    bkg_guess = np.median(y)

    # Perform the curve fitting
    popt, pcov = curve_fit(
        gaussian, x, y, p0=[amp_guess, mu_guess, sigma_guess, bkg_guess]
    )

    return popt, pcov


def pixel_list_transform(pixels: list):
    """Transform a list of pixels into two separate lists.

    Transform the given list of pixels into two separate lists,
    based on the input type (list of integers, of lists, or a mix of
    both).

    Parameters:
        pixels : list
            A list of pixels.

    Returns:
        list: A list of the left pixels.
        list: A list of the right pixels.
    """

    if isinstance(pixels[0], list) and isinstance(pixels[1], list) is True:
        pixels_left, pixels_right = sorted(pixels)
    elif isinstance(pixels[0], int) and isinstance(pixels[1], list) is True:
        pixels_left, pixels_right = sorted([[pixels[0]], pixels[1]])
    elif isinstance(pixels[0], list) and isinstance(pixels[1], int) is True:
        pixels_left, pixels_right = sorted([pixels[0], [pixels[1]]])
    elif isinstance(pixels[0], int) and isinstance(pixels[1], int) is True:
        pixels_left = [pixels[0]]
        pixels_right = [pixels[1]]

    return [pixels_left, pixels_right]


def __correct_pix_address(pix: int):
    """Pixel address correction.

    Should be used internally only with the "correct_pixel_address"
    function. Transforms the pixel address based on its position.

    Parameters
    ----------
    pix : int
        Pixel address.

    Returns
    -------
    int
        Transformed pixel address.
    """
    if pix > 127:
        pix = 255 - pix
    else:
        pix = pix + 128
    return pix


def correct_pixels_address(pixels: List[int] | List[List[int]]):
    """Correct pixel address for all given pixels.

    Return the list with the same dimensions as the input.

    Parameters
    ----------
    pixels : List[int] | List[List[int]]
        List of pixel addresses.

    Returns
    -------
    List[int] | List[List[int]]
        List of transformed pixel addresses.
    """
    if isinstance(pixels, list):
        return [correct_pixels_address(item) for item in pixels]
    else:
        return __correct_pix_address(pixels)


def file_rewrite_handling(file: str, rewrite: bool):
    """Handle file rewriting based on the 'rewrite' parameter.

    The function checks if the specified file exists in the
    'delta_ts_data' directory.
    If it exists and 'rewrite' is True, the file is deleted after a
    countdown (5 sec), this is a good time to stop execution if needed.
    If 'rewrite' is False and the file exists, the function exits with a
    system exit error message.

    Parameters
    ----------
    file : str
        The file path or name to check for existence and potentially
        rewrite.
    rewrite : bool
        If True, the function attempts to rewrite the file after
        deleting it.
        If False and the file already exists, the function exits with an
        error message.

    Returns
    -------
    None

    Raises
    ------
    SystemExit
        If 'rewrite' is False and the file already exists, the function
        exits with an error message.

    """
    try:
        # os.chdir("delta_ts_data")
        if os.path.isfile(file):
            if rewrite is True:
                print(
                    "\n! ! ! Feather file with timestamps differences already "
                    "exists and will be rewritten ! ! !\n"
                )
                for i in range(5):
                    print(f"\n! ! ! Deleting the file in {5 - i} ! ! !\n")
                    time.sleep(1)
                os.remove(file)
            else:
                sys.exit(
                    "\n Feather file already exists, 'rewrite' set to"
                    "'False', exiting."
                )
        # os.chdir("..")
    except FileNotFoundError:
        pass


def error_propagation_division(x, sigma_x, y, sigma_y, rho_xy=0):
    """Calculate error propagation for division operation.

    Parameters
    ----------
    x : float
        Numerator value.
    sigma_x : float
        Uncertainty (standard deviation) associated with x.
    y : float
        Denominator value.
    sigma_y : float
        Uncertainty (standard deviation) associated with y.
    rho_xy : float, optional
        Correlation coefficient between x and y (default is 0).

    Returns
    -------
    float
        The uncertainty (standard deviation) of the result of the division.

    Raises
    ------
    ValueError
        If any of the input values are non-numeric.
    """
    partial_derivative_x = 1 / y
    partial_derivative_y = -x / (y**2)

    term1 = (partial_derivative_x**2) * (sigma_x**2)
    term2 = (partial_derivative_y**2) * (sigma_y**2)
    term3 = (
        2
        * partial_derivative_x
        * partial_derivative_y
        * rho_xy
        * sigma_x
        * sigma_y
    )

    sigma_f = np.sqrt(term1 + term2 - term3)

    return sigma_f


def combine_feather_files(path: str, skip_data: bool = False):
    """Combine ".feather" files into one.

    Find all numbered ".feather" files for the data files found in the
    path and combine them all into one.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    skip_data : bool
        Switch for skipping the data and working directly in the
        'delta_ts_data' folder. Can be used when the raw '.dat' files
        are not available or the data set is incomplete. The default is
        False.

    Raises
    ------
    FileNotFoundError
        Raised when the folder "delta_ts_data", where timestamp
        differences are saved, cannot be found in the path.
    """

    os.chdir(path)

    if not skip_data:
        files_all = sorted(glob.glob("*.dat*"))
        if files_all != []:
            feather_file_name = files_all[0][:-4] + "-" + files_all[-1][:-4]
            combined_feather_file_name = feather_file_name
        else:
            feather_file_name = ""
            combined_feather_file_name = "combined"
    else:
        feather_file_name = ""
        combined_feather_file_name = "combined"

    try:
        os.chdir("delta_ts_data")
    except FileNotFoundError:
        raise FileNotFoundError(
            "Folder with saved timestamp differences was not found"
        )

    file_pattern = f"{feather_file_name}*_*.feather"

    feather_files = glob.glob(file_pattern)

    data_combined = []
    data_combined = pd.DataFrame(data_combined)

    for ft_file in feather_files:
        data = ft.read_feather(ft_file)

        data_combined = pd.concat([data_combined, data], ignore_index=True)

        data_combined.to_feather(f"{combined_feather_file_name}.feather")

    for ft_file in feather_files:
        os.remove(ft_file)
